#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

import tempfile
import fnmatch
from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Base import GangaObject
from Ganga.Utility.Config import getConfig, ConfigError
import Ganga.Utility.logging
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
from Ganga.GPIDev.Lib.Job.Job import Job, JobTemplate

logger = Ganga.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class GaudiDataset(Dataset):

    '''Class for handling Gaudi data sets.
    '''
    schema = {}
    docstr = 'List of PhysicalFile and DiracFile objects'
    schema['files'] = SimpleItem(defvalue=[], typelist=['str', 'type(None)'],
                                 sequence=1, doc=docstr)

    _schema = Schema(Version(3, 0), schema)
    _category = 'datasets'
    _name = "GaudiDataset"
    _exportmethods = ['__len__', '__getitem__', '__setitem__', '__delitem__',
                      '__iter__', '__reversed__', '__contains__', '__getslice__',
                      '__setslice__', '__delslice__', '__add__', '__iadd__', 'append',
                      'extend', 'insert', 'remove', 'pop', 'index', 'count', 'sort',
                      'reverse']

    def __init__(self, files=None):
        if files is None:
            files = []
        super(GaudiDataset, self).__init__()
        self.files = files

    def isEmpty(self):
        return len(self.files) == 0

    # List wrapping methods.
    def __len__(self): return self.files.__len__()

    def __getitem__(self, key): return self.files.__getitem__(key)

    def __setitem__(
        self, key, value): return self.files.__setitem__(key, value)

    def __delitem__(self, key): return self.files.__delitem__(key)

    def __iter__(self): return self.files.__iter__()

    def __reversed__(self): return self.files.__reversed__()

    def __contains__(self, item): return self.files.__contains__(item)

    def __getslice__(self, i, j): return self.files.__getslice__(i, j)

    def __setslice__(self, i, j, sequence): return self.files.__setslice__(
        i, j, sequence)

    def __delslice__(self, i, j): return self.files.__delslice__(i, j)

    def __add__(self, other): return self.files.__add__(other)

    def __iadd__(self, other): return self.files.__iadd__(other)

    def append(self, x): return self.files.append(x)

    def extend(self, L): return self.files.extend(L)

    def insert(self, i, x): return self.files.insert(i, x)

    def remove(self, x): return self.files.remove(x)

    def pop(self, *i): return self.files.pop(*i)

    def index(self, x): return self.files.index(x)

    def count(self, x): return self.files.count(x)

    @staticmethod
    def sort(self): return self.files.sort()

    @staticmethod
    def reverse(self): return self.files.reverse()


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from Ganga.GPIDev.Base.Filters import allComponentFilters


def string_dataset_shortcut(files, item):
    from Ganga.GPIDev.Base.Objects import ObjectMetaclass
    from Ganga.GPIDev.Base.PRoxy import stripProxy
    filterList = [stripProxy(i)._schema.datadict['inputdata'] for i in Ganga.GPI.__dict__.values()
                  if isinstance(stripProxy(i), ObjectMetaclass)
                  and issubclass(stripProxy(i), Job)
                  and 'inputdata' in stripProxy(i)._schema.datadict]

    # job.outputdata not used any more, use job.outputfiles
    #filterList  = [Job._schema['inputdata'], Job._schema['outputdata']]
    if type(files) is not type([]):
        return None
    if item in filterList:  # inputdataList:
        ds = GaudiDataset(files)
        return ds
    else:
        return None  # used to be c'tors, but shouldn't happen now

allComponentFilters['datasets'] = string_dataset_shortcut

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



from Ganga.GPIDev.Lib.Tasks.common import *
import random

from Ganga.GPIDev.Lib.Tasks import Transform
from TaskApplication import AthenaMCTask
from GangaAtlas.Lib.AthenaMC.AthenaMCDatasets import AthenaMCOutputDatasets, AthenaMCInputDatasets, _usertag
from Ganga.GPIDev.Schema import *
from Ganga.Utility.ColourText import overview_colours, status_colours, ANSIMarkup
markup = ANSIMarkup()

# Extract username from certificate
from Ganga.GPIDev.Credentials import GridProxy
proxy = GridProxy()
username = proxy.identity()
      
# AthenaMC public settings
#athenamcsettings = ["mode", "run_number", "production_name", "process_name", "atlas_release", "se_name", "transform_archive", "verbosity", "siteroot", "version", "cmtsite", "transform_script", "extraArgs", "extraIncArgs"]
athenamcsettings = ["mode", "run_number", "production_name", "process_name", "atlas_release", "se_name", "transform_archive"]
settings = {}
settings["EvgenTransform"] = athenamcsettings + ["evgen_job_option"]
settings["SimulTransform"] = athenamcsettings + ["triggerConfig", "geometryTag"]
settings["ReconTransform"] = athenamcsettings + ["triggerConfig", "geometryTag"]

class MCTransform(Transform):
   _schema = Schema(Version(1,0), dict(Transform._schema.datadict.items() + {
       'file_type': SimpleItem(defvalue=1, hidden=1, doc='string in the output file, pe. evgen.EVNT..',modelist=["str"]),
       'random_seeds': SimpleItem(defvalue={}, doc='random seeds to be used in the partition',modelist=["dict","int"]),

}.items()))
   _category = 'transforms'
   _name = 'MCTransform'
   _hidden = 1
   _exportmethods = Transform._exportmethods

## Special methods
   def initialize(self):
      super(MCTransform, self).initialize()
      self.application = AthenaMCTask()
      self.outputdata = AthenaMCOutputDatasets()

## Private methods (overridden from Transforms)
   def checkCompletedApp(self, app):
      if "dryrun" in stripProxy(app)._data and app.dryrun:
         return True
      j = app._getParent()
      for f in j.outputdata.actual_output:
         if self.file_type in f:
            return True
      logger.error("Job %s has not produced %s file, only: %s" % (j.fqid, self.file_type, j.outputdata.actual_output))
      return False

   def getJobsForPartitions(self, partitions):
      """Create Ganga Jobs for the next N partitions that are ready and submit them."""
      j = self.createNewJob(partitions[0])
      # Set random seed - random.seed() initializes with nanosecond time

      for part in partitions:
         random.seed()
         if not part in self.random_seeds:
            self.random_seeds[part] = random.randint(2**16,2**23)
      # Set splitter if number > 1
      if len(partitions) > 1:
         j.splitter=GPI.AthenaMCTaskSplitterJob()
         j.splitter.output_partitions = partitions
         j.splitter.task_partitions = partitions
         j.splitter.random_seeds = [self.random_seeds[part] for part in partitions]
      else:
         j.application.partition_number = partitions[0]
         j.application.random_seed = "%s" % self.random_seeds[partitions[0]]
      return [j]

   def updateInputStatus(self, ltf, partition):
      # Check all partitions that use this input
      partitions = self.application.getPartitionsForInputs([partition], self.inputdata)
      for c in partitions:
         if self.getPartitionStatus(c) in ["completed", "bad", "ignored"]:
            continue
         dep_status = [ltf.getPartitionStatus(i) for i in self.application.getInputsForPartitions([c], self.inputdata)]
         ready = [s in ["completed","bad","ignored"] for s in dep_status]
         if False in ready:
            self.setPartitionsStatus([c],"hold")
         elif "completed" in dep_status:
            self.setPartitionsStatus([c],"ready")
         else: # if all dependencies are on "bad", mark this as bad as well
            logger.warning("All input files of partition %i in %s are marked as 'bad' or 'ignored'. This partition is marked as 'bad' as well.", partition, self.fqn())
            self.setPartitionsStatus([c],"bad")

## Information methods
   def info(self):
      print markup("%s '%s'" % (self.__class__.__name__, self.name), status_colours[self.status])
      skipstring = ""
      if self.inputdata and self.inputdata.skip_files > 0:
         skipstring += "skipping " + say(self.inputdata.skip_files,"file")
         if self.inputdata.skip_events > 0:
            skipstring += " and " + say(self.inputdata.skip_events,"event")
      elif self.inputdata and self.inputdata.skip_events > 0:
         skipstring += "skipping " + say(self.inputdata.skip_events,"event")
      elif not self.inputdata:
         skipstring += "skipping " + say(self.application.firstevent-1,"event")

      print "* %s per job" % say(self.application.number_events_job,"event")
      if skipstring != "":
         print "* " + skipstring
      print "* backend: %s" % self.backend.__class__.__name__
      print "* application settings:"
      for setting in settings[self.__class__.__name__]:
         print "  - %20s: %s" % (setting, self.application._data[setting])

   def overview(self):
      if self.inputdata and self._getParent() and self._getParent().transforms.index(self) == 0:
         if not "turls" in self.inputdata.__dict__:
            self.inputdata.get_dataset(self.application, self.backend._name)
         # compare MCTask.py check() function
         partitions = self.inputdata.filesToNumbers(self.inputdata.turls.keys())
         partitions.sort()
         o = markup("Inputdataset '%s' of %s '%s':\n" % (self.inputdata.DQ2dataset, self.__class__.__name__, self.name), status_colours["running"])
         for p in range(partitions[0], partitions[-1]+1):
            if p in partitions:
               o += markup("%i " % p, overview_colours["completed"])
            else:
               o += markup("%i " % p, overview_colours["hold"])
            if p % 20 == 0: o+="\n"
         print o
      super(MCTransform, self).overview()

class EvgenTransform(MCTransform):
   _schema = Schema(Version(1,0), dict(MCTransform._schema.datadict.items() + {}.items()))
   _category = 'transforms'
   _name = 'EvgenTransform'
   _exportmethods = MCTransform._exportmethods

   def initialize(self):
      super(EvgenTransform, self).initialize()
      self.name = "Evgen"
      self.application.mode = "evgen"
      self.application.number_events_job = 10000
      self.file_type = "evgen.EVNT"

class SimulTransform(MCTransform):
   _schema = Schema(Version(1,0), dict(MCTransform._schema.datadict.items() + {}.items()))
   _category = 'transforms'
   _name = 'SimulTransform'
   _exportmethods = MCTransform._exportmethods

   def initialize(self):
      super(SimulTransform, self).initialize()
      self.name = "Simul"
      self.application.number_events_job = 50
      self.application.mode = "simul"
      self.file_type = "simul.RDO"
      self.inputdata = AthenaMCInputDatasets()
      self.inputdata.use_partition_numbers = True

   def getJobsForPartitions(self, partitions):
      jl = super(SimulTransform, self).getJobsForPartitions(partitions)
      return jl

class ReconTransform(MCTransform):
   _schema = Schema(Version(1,0), dict(MCTransform._schema.datadict.items() + {}.items()))
   _category = 'transforms'
   _name = 'ReconTransform'
   _exportmethods = MCTransform._exportmethods

   def initialize(self):
      super(ReconTransform, self).initialize()
      self.name = "Recon"
      self.application.number_events_job = 1000
      self.application.mode = "recon"
      self.file_type = "recon.AOD"
      self.inputdata = AthenaMCInputDatasets()
      self.inputdata.use_partition_numbers = True




######################################################
# RATUser.py
# ---------
# Author: Matt Mottram
#         <m.mottram@qmul.ac.uk>
#
# Description:
#    Ganga application for SNO+ user analysis/simulation.
#
# Runs RAT snapshots on the given backend via ratRunner.py
# Ships the code in the input sandbox, can either download
# the snapshot from a given rat fork, or checkout and tar
# up the snapshot from a local repository.
#
# Classes:
#  - RATUser: user analysis/simulation applicaiton
#  - RATUserSplitter: splitter to create subjobs for subruns
#  - UserRTHandler: handles submission to local/batch backends
#  - UserLCGRTHandler: handles submission to LCG backend
#
# Revision History:
#  - 03/11/12: M. Mottram: first revision with proper documentation!
#  - 06/12/13: M. Mottram: Removed use of tokens, updated config and schema.
#  - 26/08/14: M. Mottram: moved RT handlers to separate file (RTRATUser.py)
# 
######################################################

import os

from GangaSNOplus.Lib.Utilities import RATUtil

from Ganga.Core.exceptions import ApplicationConfigurationError

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Schema import *

from Ganga.GPIDev.Lib.File import *

###################################################################

config = Ganga.Utility.Config.makeConfig('defaults_RATUser','Defaults for the RATUser application')

config.addOption('local_softwareEnvironment', None, 'Local snoing-install directory (or directory with env_rat-x.y.sh files)')
config.addOption('local_environment', [], 'Environment options required to run on local or batch system')
config.addOption('local_outputDir', None, '*Default* output directory if running on a batch or local system (can override)')
config.addOption('grid_outputDir', None, '*Defult* output directory if running on system with grid storage (can override)')
config.addOption('cacheDir', '~/gaspCache', 'Directory to store RAT snaphots (if required)')


# Assume that the applications should come from the same GangaSNOplus directory        
_app_directory = os.path.dirname(__file__)

###################################################################

class RATUser(IApplication):
    """The RAT job handler for data production and processing"""

    #_schema is required for any Ganga plugin
    #Add any options that are required, but try to set sensible default values to minimise effort required of user
    _schema = Schema(Version(1,1), {
            'discardOutput'     : SimpleItem(defvalue=False,doc='Do not store the output: default False',typelist=['bool']),
            'environment'       : SimpleItem(defvalue=[],doc='list of strings with the commands to setup the correct backend environment, or single string location of a file with the appropriate commands (if necessary)',typelist=['str','list']),
            'inputFile'         : SimpleItem(defvalue=None,doc='Input file name, macro cannot have the inroot process defined within!',
                                             typelist=['str','type(None)']),
            'nEvents'           : SimpleItem(defvalue=None,doc='Number of events to run, MUST not define number of events in the macro (/rat/run/start)',
                                             typelist=['int','type(None)']),
            'outputDir'         : SimpleItem(defvalue=None,doc='Which output directory should we use (default Grid: RATUser/general/, should be modified to RATUser/<your-name> if you dont want admins to mess with it!)',
                                             typelist=['str','type(None)']),
            'outputFile'        : SimpleItem(defvalue=None,doc='Output file name, macro must have outroot processor, but no output file defined!',
                                             typelist=['str','type(None)']),
            'rat_db_name'       : SimpleItem(defvalue=None, doc='RAT db name', typelist=['str', 'type(None)']),
            'rat_db_pswd'       : SimpleItem(defvalue=None, doc='RAT db password', typelist=['str', 'type(None)']),
            'rat_db_protocol'   : SimpleItem(defvalue=None, doc='RAT db protocol', typelist=['str', 'type(None)']),
            'rat_db_url'        : SimpleItem(defvalue=None, doc='RAT db password', typelist=['str', 'type(None)']),
            'rat_db_user'       : SimpleItem(defvalue=None, doc='RAT db password', typelist=['str', 'type(None)']),
            'ratBaseVersion'    : SimpleItem(defvalue='dev',doc='RAT version that ratVersion derives from, necessary to get the correct libraries (ROOT, Geant4 etc)',
                                             typelist=['str',"int"]),
            'ratFork'           : SimpleItem(defvalue='snoplus', doc='Fork of RAT [snoplus]', typelist=['str']),
            'ratMacro'          : SimpleItem(defvalue=None,doc='String pointing to the macro file to run',
                                             typelist=['str','type(None)']),
            'ratVersion'        : SimpleItem(defvalue=None,doc='RAT version tag for the version to download and install (can also be a branch name, not recommended)',
                                             typelist=['str','type(None)']),
            'softwareEnvironment': SimpleItem(defvalue=None,doc='Software environment file, required if running on a non-LCG backend',
                                             typelist=['str','type(None)']),
            'tRun'              : SimpleItem(defvalue=None,doc='Duration of run (cannot use with nEvents)',
                                             typelist=['int','type(None)']),
            'useDB'             : SimpleItem(defvalue=False,doc='Use the RAT database (snopl.us)?',typelist=['bool']),
            'versionUpdate'     : SimpleItem(defvalue=False, doc="Update the rat version tag?", typelist=['bool']),
            })
    
    _category = 'applications'
    _name = 'RATUser'

    config = Ganga.Utility.Config.getConfig('defaults_RATUser')
    def configure(self,masterappconfig):
        '''Configure method, called once per job.
        '''
        logger.debug('RAT::RATUser configure ...')

        job = self._getParent()
        masterjob = job._getParent()

        #Critical options:
        # - ratMacro 
        # - outputDirectory
        # - ratBaseVersion
        #If these aren't defined, don't let the user submit the job
        #Note, the ratMacro can be defined in the subjob...
        if self.ratMacro!=None:
            job.inputsandbox.append(File(self.ratMacro))
        else:
            logger.error('Rat macro not defined')
            raise Exception
        if self.ratBaseVersion=='':
            logger.error('Error: must give a rat base (fixed release) version number')
            raise Exception
        if not self.outputFile:
            logger.error('No output file defined!')   #checks if output file set in command line
            raise Exception 
        if RATUtil.check_command(self.ratMacro,['/rat/procset','file']):
            logger.error('Output file should only be defined via the command line and not in the macro!')  #checks if output file si set in the macro
            raise Exception
        if self.outputFile:
            if not RATUtil.check_command(self.ratMacro,['/rat/proclast','outroot']) and not RATUtil.check_command(self.ratMacro,['/rat/proclast','outntuple']) and not RATUtil.check_command(self.ratMacro,['/rat/proclast','outsoc']):
                logger.error('Have specified an output file, but no root, ntuple or soc processor present in macro')  #checks for putroot processor (needs to be there regardless where the output file is defined)
                raise Exception
        if not self.nEvents and not self.tRun:
            logger.error('Need to specifiy either the number of events or the duration of run! No number of events or run duration defined!') #checks if number of events or run duration is set in command line
            raise Exception
        if self.nEvents and self.tRun:
            logger.error('Cannot specify number of events and the duration of run!')
            raise Exception
        if not RATUtil.check_command(self.ratMacro,['/rat/run/start','']):
            logger.error('/rat/run/start must be set in the macro but no number of events should be specified! Number of events should only be defined via the command line!')  #check if the /rat/run/start command is set
            raise Exception
        if self.inputFile:
            if RATUtil.checkText(self.ratMacro,['inroot/read']):
                logger.error('Cannot specify inputFile in Ganga job if "/rat/inroot/read" line is present in macro')
                raise Exception
        if self.useDB:
            if not config['rat_db_pswd']:
                logger.error('Need a password in order to contact the ratdb database')
                raise Exception

        #Always run rat with a log called rat.log
        job.outputsandbox.append('rat.log')
        job.outputsandbox.append('return_card.js')

        if self.ratVersion!=None:
            #download the code locally
            #only uses the main SNO+ rat branch for now
            #need to add pkl object to inform which branch we have and add others when required
            self.zipFileName = RATUtil.make_rat_snapshot(self.ratFork, self.ratVersion, self.versionUpdate, os.path.expanduser(config['cacheDir']))
            job.inputsandbox.append(File(self.zipFileName))

        #all args have to be str/file - force rat base version to be a string
        self.ratBaseVersion=str(self.ratBaseVersion)

        return(None,None)
                                
###################################################################

class RATUserSplitter(ISplitter):
    '''Splitter for RAT User jobs.
    '''
    _name = "RATUserSplitter"
    _schema = Schema(Version(1,0), {
            'ratMacro' : SimpleItem(defvalue=[],typelist=['str'],sequence=1,doc='A list of lists for specifying rat macros files'),
            'outputFile' : SimpleItem(defvalue=[],typelist=['str','type(None)'],sequence=1,doc='A list of lists for specifying rat output files'),
            'inputFile' : SimpleItem(defvalue=[],typelist=['str','type(None)'],sequence=1,doc='A list of lists for specifying rat input files'),
            'nEvents' : SimpleItem(defvalue=[],typelist=['int','type(None)'],sequence=1,doc='A list of the number of events for each sub job')
        } )

    def split(self,job):

        if self.outputFile!=[]:
            if len(self.outputFile)!=len(self.ratMacro):
                logger.error('Must have same number of macros, outputs and inputs for the splitter')
                raise Exception
        if self.inputFile!=[]:
            if len(self.inputFile)!=len(self.ratMacro):
                logger.error('Must have same number of macros, outputs and inputs for the splitter')
                raise Exception
        if self.nEvents!=[]:
            if len(self.nEvents)!=len(self.ratMacro):
                logger.error('Must have same number of nEvents as macros for the splitter')
                raise Exception
        
        subjobs = []

        for i,rm in enumerate(self.ratMacro):
            j = self.createSubjob(job)
            j.application.ratMacro = rm
            if self.outputFile!=[]:
                j.application.outputFile = self.outputFile[i]
            if self.inputFile!=[]:
                j.application.inputFile = self.inputFile[i]
            if self.nEvents!=[]:
                j.application.nEvents = self.nEvents[i]
            subjobs.append(j)
        return subjobs
    
###################################################################

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaSNOplus.Lib.RTHandlers.RTRATUser import UserRTHandler, UserLCGRTHandler, UserWGRTHandler, UserDiracRTHandler

allHandlers.add('RATUser','Local', UserRTHandler)
allHandlers.add('RATUser','PBS', UserRTHandler)
allHandlers.add('RATUser','SGE', UserRTHandler)
allHandlers.add('RATUser','Condor', UserRTHandler)
allHandlers.add('RATUser','LCG', UserLCGRTHandler)
allHandlers.add('RATUser','Dirac', UserDiracRTHandler)
allHandlers.add('RATUser','TestSubmitter', UserRTHandler)
allHandlers.add('RATUser','Interactive', UserRTHandler)
allHandlers.add('RATUser','Batch', UserRTHandler)
allHandlers.add('RATUser','WestGrid', UserWGRTHandler)

logger = Ganga.Utility.logging.getLogger()



from __future__ import absolute_import
from .Transform import Transform
from Ganga.Core import ApplicationConfigurationError
from Ganga.GPIDev.Schema import Schema
from Ganga.GPIDev.Schema import SimpleItem
from Ganga.GPIDev.Schema import Version
from Ganga.Utility.util import containsGangaObjects
from .TaskApplication import ArgSplitterTask


class ArgTransform(Transform):
    _schema = Schema(Version(1, 0), dict(Transform._schema.datadict.items() + {
        'args': SimpleItem(defvalue=[], typelist=['list', 'Ganga.GPIDev.Lib.GangaList.GangaList.GangaList'], sequence=1, doc='A list of lists of arguments to pass to script'),
    }.items()))
    _category = 'transforms'
    _name = 'ArgTransform'
    _exportmethods = Transform._exportmethods

    def check(self):
        nargs = len(self.args)
        self.setPartitionsStatus(range(1, nargs + 1), "ready")
        if "_partition_status" in self.getNodeData():
            self.setPartitionsLimit(nargs + 1)

    def getJobsForPartitions(self, partitions):
        """Create Ganga Jobs for the next N partitions that are ready and submit them."""
        j = self.createNewJob(partitions[0])
        if len(partitions) > 1:
            j.splitter = ArgSplitterTask()
            j.splitter.args = [self.args[p - 1] for p in partitions]
            j.splitter.task_partitions = partitions
        else:
            p = partitions[0]
            if (p < 1 or p > len(self.args)):
                raise ApplicationConfigurationError(
                    "Partition %i did not find a corresponding argment!", p)
            j.application.args = self.args[p - 1]
        return [j]



from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Schema import Schema, SimpleItem, Version
from Ganga.Core.exceptions import SplitterError
from xml.dom.minidom import parse

import os


class CRABSplitter(ISplitter):
    """Splitter object for CRAB jobs."""
    schemadic = {}
    schemadic['maxevents'] = SimpleItem(defvalue=None,
                                        typelist=['type(None)', 'int'],
                                        doc='Maximum number of events/task')
    schemadic['inputfiles'] = SimpleItem(defvalue=None,
                                         typelist=['type(None)', 'str'],
                                         doc='Number of input files')
    schemadic['skipevents'] = SimpleItem(defvalue=None,
                                         typelist=['type(None)', 'int'],
                                         doc='Offset for the events')
    _name = 'CRABSplitter'
    _schema = Schema(Version(1, 0), schemadic)

    def parseArguments(self, path):
        """Gets some job arguments from the FJR."""
        splittingData = []
        for job in parse(path).getElementsByTagName("Job"):
            splittingData.append([job.getAttribute("MaxEvents"),
                                  job.getAttribute("InputFiles"),
                                  job.getAttribute("SkipEvents")])
        return splittingData

    def split(self, job):
        """Main splitter for the job."""
        try:
            arguments_path = os.path.join(job.inputdata.ui_working_dir,
                                          'share/arguments.xml')
            splittingData = self.parseArguments(arguments_path)
        except IOError as e:
            raise SplitterError(e)

        subjobs = []
        for maxevents, inputfiles, skipevents in splittingData:
            j = self.createSubjob(job)
            j.master = job
            j.application = job.application
            j.inputdata = job.inputdata
            j.backend = job.backend

            splitter = CRABSplitter()
            splitter.maxevents = maxevents
            splitter.inputfiles = inputfiles
            splitter.skipevents = skipevents
            j.splitter = splitter
            subjobs.append(j)
        return subjobs



# ARC backend
import os
import os.path
import math
import re

from urlparse import urlparse

from Ganga.Core.GangaThread.MTRunner import MTRunner, Data, Algorithm
from Ganga.Core import GangaException

from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from Ganga.GPIDev.Lib.File import FileBuffer
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger, log_user_exception
from Ganga.Lib.LCG.Utility import get_md5sum
from Ganga.Lib.LCG.ElapsedTimeProfiler import ElapsedTimeProfiler

from Ganga.Lib.LCG.Grid import Grid
from Ganga.Lib.LCG.LCG import grids
from Ganga.Lib.LCG.GridftpSandboxCache import GridftpSandboxCache

from Ganga.GPIDev.Base.Proxy import getName

class ARC(IBackend):

    '''ARC backend - direct job submission to an ARC CE'''
    _schema = Schema(Version(1, 0), {
        'CE': SimpleItem(defvalue='', doc='ARC CE endpoint'),
        'jobtype': SimpleItem(defvalue='Normal', doc='Job type: Normal, MPICH'),
        'requirements': ComponentItem('LCGRequirements', doc='Requirements for the resource selection'),
        'sandboxcache': ComponentItem('GridSandboxCache', copyable=1, doc='Interface for handling oversized input sandbox'),
        'id': SimpleItem(defvalue='', typelist=['str', 'list'], protected=1, copyable=0, doc='Middleware job identifier'),
        'status': SimpleItem(defvalue='', typelist=['str', 'dict'], protected=1, copyable=0, doc='Middleware job status'),
        'exitcode': SimpleItem(defvalue='', protected=1, copyable=0, doc='Application exit code'),
        'exitcode_arc': SimpleItem(defvalue='', protected=1, copyable=0, doc='Middleware exit code'),
        'actualCE': SimpleItem(defvalue='', protected=1, copyable=0, doc='The ARC CE where the job actually runs.'),
        'reason': SimpleItem(defvalue='', protected=1, copyable=0, doc='Reason of causing the job status'),
        'workernode': SimpleItem(defvalue='', protected=1, copyable=0, doc='The worker node on which the job actually runs.'),
        'isbURI': SimpleItem(defvalue='', protected=1, copyable=0, doc='The input sandbox URI on ARC CE'),
        'osbURI': SimpleItem(defvalue='', protected=1, copyable=0, doc='The output sandbox URI on ARC CE'),
        'verbose': SimpleItem(defvalue=False, doc='Use verbose options for ARC commands')
    })

    _category = 'backends'

    _name = 'ARC'

    def __init__(self):
        super(ARC, self).__init__()

        # dynamic requirement object loading
        try:
            reqName1 = config['Requirements']
            reqName = config['Requirements'].split('.').pop()
            reqModule = __import__(reqName1, globals(), locals(), [reqName1])
            reqClass = vars(reqModule)[reqName]
            self.requirements = reqClass()

            logger.debug('load %s as LCGRequirements' % reqName)
        except:
            logger.debug('load default LCGRequirements')
            pass

        # dynamic sandbox cache object loading
        # force to use GridftpSandboxCache
        self.sandboxcache = GridftpSandboxCache()
        try:
            scName1 = config['SandboxCache']
            scName = config['SandboxCache'].split('.').pop()
            scModule = __import__(scName1, globals(), locals(), [scName1])
            scClass = vars(scModule)[scName]
            self.sandboxcache = scClass()
            logger.debug('load %s as SandboxCache' % scName)
        except:
            logger.debug('load default SandboxCache')
            pass

    def __refresh_jobinfo__(self, job):
        '''Refresh the lcg jobinfo. It will be called after resubmission.'''
        job.backend.status = ''
        job.backend.reason = ''
        job.backend.actualCE = ''
        job.backend.exitcode = ''
        job.backend.exitcode_arc = ''
        job.backend.workernode = ''
        job.backend.isbURI = ''
        job.backend.osbURI = ''

    def __setup_sandboxcache__(self, job):
        '''Sets up the sandbox cache object to adopt the runtime configuration of the LCG backend'''

        re_token = re.compile('^token:(.*):(.*)$')

        self.sandboxcache.vo = config['VirtualOrganisation']
        self.sandboxcache.middleware = 'GLITE'
        self.sandboxcache.timeout = config['SandboxTransferTimeout']

        if self.sandboxcache._name == 'LCGSandboxCache':
            if not self.sandboxcache.lfc_host:
                self.sandboxcache.lfc_host = grids[
                    self.sandboxcache.middleware].__get_lfc_host__()

            if not self.sandboxcache.se:

                token = ''
                se_host = config['DefaultSE']
                m = re_token.match(se_host)
                if m:
                    token = m.group(1)
                    se_host = m.group(2)

                self.sandboxcache.se = se_host

                if token:
                    self.sandboxcache.srm_token = token

            if (self.sandboxcache.se_type in ['srmv2']) and (not self.sandboxcache.srm_token):
                self.sandboxcache.srm_token = config['DefaultSRMToken']

        return True

    def __check_and_prestage_inputfile__(self, file):
        '''Checks the given input file size and if it's size is
           over "BoundSandboxLimit", prestage it to a grid SE.

           The argument is a path of the local file.

           It returns a dictionary containing information to refer to the file:

               idx = {'lfc_host': lfc_host,
                      'local': [the local file pathes],
                      'remote': {'fname1': 'remote index1', 'fname2': 'remote index2', ... }
                     }

           If prestaging failed, None object is returned.

           If the file has been previously uploaded (according to md5sum),
           the prestaging is ignored and index to the previously uploaded file
           is returned.
           '''

        idx = {'lfc_host': '', 'local': [], 'remote': {}}

        job = self.getJobObject()

        # read-in the previously uploaded files
        uploadedFiles = []

        # getting the uploaded file list from the master job
        if job.master:
            uploadedFiles += job.master.backend.sandboxcache.get_cached_files()

        # set and get the $LFC_HOST for uploading oversized sandbox
        self.__setup_sandboxcache__(job)

        uploadedFiles += self.sandboxcache.get_cached_files()

        lfc_host = None

        # for LCGSandboxCache, take the one specified in the sansboxcache object.
        # the value is exactly the same as the one from the local grid shell env. if
        # it is not specified exclusively.
        if self.sandboxcache._name == 'LCGSandboxCache':
            lfc_host = self.sandboxcache.lfc_host

        # or in general, query it from the Grid object
        if not lfc_host:
            lfc_host = grids[
                self.sandboxcache.middleware.upper()].__get_lfc_host__()

        idx['lfc_host'] = lfc_host

        abspath = os.path.abspath(file)
        fsize = os.path.getsize(abspath)

        if fsize > config['BoundSandboxLimit']:

            md5sum = get_md5sum(abspath, ignoreGzipTimestamp=True)

            doUpload = True
            for uf in uploadedFiles:
                if uf.md5sum == md5sum:
                    # the same file has been uploaded to the iocache
                    idx['remote'][os.path.basename(file)] = uf.id
                    doUpload = False
                    break

            if doUpload:

                logger.warning(
                    'The size of %s is larger than the sandbox limit (%d byte). Please wait while pre-staging ...' % (file, config['BoundSandboxLimit']))

                if self.sandboxcache.upload([abspath]):
                    remote_sandbox = self.sandboxcache.get_cached_files()[-1]
                    idx['remote'][remote_sandbox.name] = remote_sandbox.id
                else:
                    logger.error(
                        'Oversized sandbox not successfully pre-staged')
                    return None
        else:
            idx['local'].append(abspath)

        return idx

    def __mt_job_prepare__(self, rjobs, subjobconfigs, masterjobconfig):
        '''preparing jobs in multiple threads'''

        logger.warning(
            'preparing %d subjobs ... it may take a while' % len(rjobs))

        # prepare the master job (i.e. create shared inputsandbox, etc.)
        master_input_sandbox = IBackend.master_prepare(self, masterjobconfig)

        # uploading the master job if it's over the WMS sandbox limitation
        for f in master_input_sandbox:
            master_input_idx = self.__check_and_prestage_inputfile__(f)

            if not master_input_idx:
                logger.error('master input sandbox perparation failed: %s' % f)
                return None

        # the algorithm for preparing a single bulk job
        class MyAlgorithm(Algorithm):

            def __init__(self):
                Algorithm.__init__(self)

            def process(self, sj_info):
                my_sc = sj_info[0]
                my_sj = sj_info[1]

                try:
                    logger.debug("preparing job %s" % my_sj.getFQID('.'))
                    jdlpath = my_sj.backend.preparejob(
                        my_sc, master_input_sandbox)

                    if (not jdlpath) or (not os.path.exists(jdlpath)):
                        raise GangaException(
                            'job %s not properly prepared' % my_sj.getFQID('.'))

                    self.__appendResult__(my_sj.id, jdlpath)
                    return True
                except Exception as x:
                    log_user_exception()
                    return False

        mt_data = []
        for sc, sj in zip(subjobconfigs, rjobs):
            mt_data.append([sc, sj])

        myAlg = MyAlgorithm()
        myData = Data(collection=mt_data)

        runner = MTRunner(
            name='lcg_jprepare', algorithm=myAlg, data=myData, numThread=10)
        runner.start()
        runner.join(-1)

        if len(runner.getDoneList()) < len(mt_data):
            return None
        else:
            # return a JDL file dictionary with subjob ids as keys, JDL file
            # paths as values
            return runner.getResults()

    def __mt_bulk_submit__(self, node_jdls):
        '''submitting jobs in multiple threads'''

        job = self.getJobObject()

        logger.warning(
            'submitting %d subjobs ... it may take a while' % len(node_jdls))

        # the algorithm for submitting a single bulk job
        class MyAlgorithm(Algorithm):

            def __init__(self, gridObj, masterInputWorkspace, ce, arcverbose):
                Algorithm.__init__(self)
                self.inpw = masterInputWorkspace
                self.gridObj = gridObj
                self.ce = ce
                self.arcverbose = arcverbose

            def process(self, jdl_info):
                my_sj_id = jdl_info[0]
                my_sj_jdl = jdl_info[1]

                #my_sj_jid = self.gridObj.arc_submit(my_sj_jdl, self.ce, self.verbose)
                my_sj_jid = self.gridObj.arc_submit(
                    my_sj_jdl, self.ce, self.arcverbose)

                if not my_sj_jid:
                    return False
                else:
                    self.__appendResult__(my_sj_id, my_sj_jid)
                    return True

        mt_data = []
        for id, jdl in node_jdls.items():
            mt_data.append((id, jdl))

        myAlg = MyAlgorithm(gridObj=grids['GLITE'], masterInputWorkspace=job.getInputWorkspace(
        ), ce=self.CE, arcverbose=self.verbose)
        myData = Data(collection=mt_data)

        runner = MTRunner(name='arc_jsubmit', algorithm=myAlg,
                          data=myData, numThread=config['SubmissionThread'])
        runner.start()
        runner.join(timeout=-1)

        if len(runner.getDoneList()) < len(mt_data):
            # not all bulk jobs are successfully submitted. canceling the
            # submitted jobs on WMS immediately
            logger.error(
                'some bulk jobs not successfully (re)submitted, canceling submitted jobs on WMS')
            grids['GLITE'].arc_cancelMultiple(runner.getResults().values())
            return None
        else:
            return runner.getResults()

    def __jobWrapperTemplate__(self):
        '''Create job wrapper'''

        script = """#!/usr/bin/env python
#-----------------------------------------------------
# This job wrapper script is automatically created by
# GANGA LCG backend handler.
#
# It controls:
# 1. unpack input sandbox
# 2. invoke application executable
# 3. invoke monitoring client
#-----------------------------------------------------
import os,os.path,shutil,tempfile
import sys,popen2,time,traceback

#bugfix #36178: subprocess.py crashes if python 2.5 is used
#try to import subprocess from local python installation before an
#import from PYTHON_DIR is attempted some time later
try:
    import subprocess
except ImportError:
    pass

## Utility functions ##
def timeString():
    return time.strftime('%a %b %d %H:%M:%S %Y',time.gmtime(time.time()))

def printInfo(s):
    out.write(timeString() + '  [Info]' +  ' ' + str(s) + os.linesep)
    out.flush()

def printError(s):
    out.write(timeString() + ' [Error]' +  ' ' + str(s) + os.linesep)
    out.flush()

def lcg_file_download(vo,guid,localFilePath,timeout=60,maxRetry=3):
    cmd = 'lcg-cp -t %d --vo %s %s file://%s' % (timeout,vo,guid,localFilePath)

    printInfo('LFC_HOST set to %s' % os.environ['LFC_HOST'])
    printInfo('lcg-cp timeout: %d' % timeout)

    i         = 0
    rc        = 0
    isDone    = False
    try_again = True

    while try_again:
        i = i + 1
        try:
            ps = os.popen(cmd)
            status = ps.close()

            if not status:
                isDone = True
                printInfo('File %s download from iocache' % os.path.basename(localFilePath))
            else:
                raise IOError("Download file %s from iocache failed with error code: %d, trial %d." % (os.path.basename(localFilePath), status, i))

        except IOError as e:
            isDone = False
            printError(str(e))

        if isDone:
            try_again = False
        elif i == maxRetry:
            try_again = False
        else:
            try_again = True

    return isDone

## system command executor with subprocess
def execSyscmdSubprocess(cmd, wdir=os.getcwd()):

    import os, subprocess

    global exitcode

    outfile   = file('stdout','w')
    errorfile = file('stderr','w')

    try:
        child = subprocess.Popen(cmd, cwd=wdir, shell=True, stdout=outfile, stderr=errorfile)

        while 1:
            exitcode = child.poll()
            if exitcode is not None:
                break
            else:
                outfile.flush()
                errorfile.flush()
                time.sleep(0.3)
    finally:
        pass

    outfile.flush()
    errorfile.flush()
    outfile.close()
    errorfile.close()

    return True

## system command executor with multi-thread
## stderr/stdout handler
def execSyscmdEnhanced(cmd, wdir=os.getcwd()):

    import os, threading

    cwd = os.getcwd()

    isDone = False

    try:
        ## change to the working directory
        os.chdir(wdir)

        child = popen2.Popen3(cmd,1)
        child.tochild.close() # don't need stdin

        class PipeThread(threading.Thread):

            def __init__(self,infile,outfile,stopcb):
                self.outfile = outfile
                self.infile = infile
                self.stopcb = stopcb
                self.finished = 0
                threading.Thread.__init__(self)

            def run(self):
                stop = False
                while not stop:
                    buf = self.infile.read(10000)
                    self.outfile.write(buf)
                    self.outfile.flush()
                    time.sleep(0.01)
                    stop = self.stopcb()
                #FIXME: should we do here?: self.infile.read()
                #FIXME: this is to make sure that all the output is read (if more than buffer size of output was produced)
                self.finished = 1

        def stopcb(poll=False):
            global exitcode
            if poll:
                exitcode = child.poll()
            return exitcode != -1

        out_thread = PipeThread(child.fromchild, sys.stdout, stopcb)
        err_thread = PipeThread(child.childerr, sys.stderr, stopcb)

        out_thread.start()
        err_thread.start()
        while not out_thread.finished and not err_thread.finished:
            stopcb(True)
            time.sleep(0.3)

        sys.stdout.flush()
        sys.stderr.flush()

        isDone = True

    except(Exception,e):
        isDone = False

    ## return to the original directory
    os.chdir(cwd)

    return isDone

############################################################################################

###INLINEMODULES###

############################################################################################

## Main program ##

outputsandbox = ###OUTPUTSANDBOX###
input_sandbox = ###INPUTSANDBOX###
wrapperlog = ###WRAPPERLOG###
appexec = ###APPLICATIONEXEC###
appargs = ###APPLICATIONARGS###
appenvs = ###APPLICATIONENVS###
timeout = ###TRANSFERTIMEOUT###

exitcode=-1

import sys, stat, os, os.path, commands

# Change to scratch directory if provided
scratchdir = ''
tmpdir = ''

orig_wdir = os.getcwd()

# prepare log file for job wrapper
out = open(os.path.join(orig_wdir, wrapperlog),'w')

if os.getenv('EDG_WL_SCRATCH'):
    scratchdir = os.getenv('EDG_WL_SCRATCH')
elif os.getenv('TMPDIR'):
    scratchdir = os.getenv('TMPDIR')

if scratchdir:
    (status, tmpdir) = commands.getstatusoutput('mktemp -d %s/gangajob_XXXXXXXX' % (scratchdir))
    if status == 0:
        os.chdir(tmpdir)
    else:
        ## if status != 0, tmpdir should contains error message so print it to stderr
        printError('Error making ganga job scratch dir: %s' % tmpdir)
        printInfo('Unable to create ganga job scratch dir in %s. Run directly in: %s' % ( scratchdir, os.getcwd() ) )

        ## reset scratchdir and tmpdir to disable the usage of Ganga scratch dir
        scratchdir = ''
        tmpdir = ''

wdir = os.getcwd()

if scratchdir:
    printInfo('Changed working directory to scratch directory %s' % tmpdir)
    try:
        os.system("ln -s %s %s" % (os.path.join(orig_wdir, 'stdout'), os.path.join(wdir, 'stdout')))
        os.system("ln -s %s %s" % (os.path.join(orig_wdir, 'stderr'), os.path.join(wdir, 'stderr')))
    except Exception as e:
        printError(sys.exc_info()[0])
        printError(sys.exc_info()[1])
        str_traceback = traceback.format_tb(sys.exc_info()[2])
        for str_tb in str_traceback:
            printError(str_tb)
        printInfo('Linking stdout & stderr to original directory failed. Looking at stdout during job run may not be possible')

os.environ['PATH'] = '.:'+os.environ['PATH']

vo = os.environ['GANGA_LCG_VO']

try:
    printInfo('Job Wrapper start.')

#   download inputsandbox from remote cache
    for f,guid in input_sandbox['remote'].iteritems():
        if not lcg_file_download(vo, guid, os.path.join(wdir,f), timeout=int(timeout)):
            raise IOError('Download remote input %s:%s failed.' % (guid,f) )
        else:
            getPackedInputSandbox(f)

    printInfo('Download inputsandbox from iocache passed.')

#   unpack inputsandbox from wdir
    for f in input_sandbox['local']:
        getPackedInputSandbox(os.path.join(orig_wdir,f))

    printInfo('Unpack inputsandbox passed.')

    #get input files
    ###DOWNLOADINPUTFILES###

    printInfo('Loading Python modules ...')

    sys.path.insert(0,os.path.join(wdir,PYTHON_DIR))

    # check the python library path
    try:
        printInfo(' ** PYTHON_DIR: %s' % os.environ['PYTHON_DIR'])
    except KeyError:
        pass

    try:
        printInfo(' ** PYTHONPATH: %s' % os.environ['PYTHONPATH'])
    except KeyError:
        pass

    for lib_path in sys.path:
        printInfo(' ** sys.path: %s' % lib_path)

#   execute application

    ## convern appenvs into environment setup script to be 'sourced' before executing the user executable

    printInfo('Prepare environment variables for application executable')

    env_setup_script = os.path.join(os.getcwd(), '__ganga_lcg_env__.sh')

    f = open( env_setup_script, 'w')
    f.write('#!/bin/sh' + os.linesep )
    f.write('##user application environmet setup script generated by Ganga job wrapper' + os.linesep)
    for k,v in appenvs.items():

        str_env = 'export %s="%s"' % (k, v)

        printInfo(' ** ' + str_env)
        
        f.write(str_env + os.linesep)
    f.close()

    try: #try to make shipped executable executable
        os.chmod('%s/%s'% (wdir,appexec),stat.S_IXUSR|stat.S_IRUSR|stat.S_IWUSR)
    except:
        pass

    status = False
    try:
        # use subprocess to run the user's application if the module is available on the worker node
        import subprocess
        printInfo('Load application executable with subprocess module')
        status = execSyscmdSubprocess('source %s; %s %s' % (env_setup_script, appexec, appargs), wdir)
    except ImportError as err:
        # otherwise, use separate threads to control process IO pipes
        printInfo('Load application executable with separate threads')
        status = execSyscmdEnhanced('source %s; %s %s' % (env_setup_script, appexec, appargs), wdir)

    os.system("cp %s/stdout stdout.1" % orig_wdir)
    os.system("cp %s/stderr stderr.1" % orig_wdir)

    printInfo('GZipping stdout and stderr...')

    os.system("gzip stdout.1 stderr.1")

    # move them to the original wdir so they can be picked up
    os.system("mv stdout.1.gz %s/stdout.gz" % orig_wdir)
    os.system("mv stderr.1.gz %s/stderr.gz" % orig_wdir)

    if not status:
        raise OSError('Application execution failed.')
    printInfo('Application execution passed with exit code %d.' % exitcode)      

    ###OUTPUTUPLOADSPOSTPROCESSING###

    for f in os.listdir(os.getcwd()):
        command = "cp %s %s" % (os.path.join(os.getcwd(),f), os.path.join(orig_wdir,f))
        os.system(command)            

    createPackedOutputSandbox(outputsandbox,None,orig_wdir)

#   pack outputsandbox
#    printInfo('== check output ==')
#    for line in os.popen('pwd; ls -l').readlines():
#        printInfo(line)

    printInfo('Pack outputsandbox passed.')

    # Clean up after us - All log files and packed outputsandbox should be in "wdir"
    if scratchdir:
        os.chdir(orig_wdir)
        os.system("rm %s -rf" % wdir)
except Exception as e:
    printError(sys.exc_info()[0])
    printError(sys.exc_info()[1])
    str_traceback = traceback.format_tb(sys.exc_info()[2])
    for str_tb in str_traceback:
        printError(str_tb)

printInfo('Job Wrapper stop.')

out.close()

# always return exit code 0 so the in the case of application failure
# one can always get stdout and stderr back to the UI for debug.
sys.exit(0)
"""
        return script

    def preparejob(self, jobconfig, master_job_sandbox):
        '''Prepare the JDL'''

        script = self.__jobWrapperTemplate__()

        job = self.getJobObject()
        inpw = job.getInputWorkspace()

        wrapperlog = '__jobscript__.log'

        import Ganga.Core.Sandbox as Sandbox

        # FIXME: check what happens if 'stdout','stderr' are specified here
        script = script.replace(
            '###OUTPUTSANDBOX###', repr(jobconfig.outputbox))

        script = script.replace(
            '###APPLICATION_NAME###', getName(job.application))
        script = script.replace(
            '###APPLICATIONEXEC###', repr(jobconfig.getExeString()))
        script = script.replace(
            '###APPLICATIONARGS###', repr(jobconfig.getArguments()))

        from Ganga.GPIDev.Lib.File.OutputFileManager import getWNCodeForOutputPostprocessing, getWNCodeForDownloadingInputFiles

        script = script.replace(
            '###OUTPUTUPLOADSPOSTPROCESSING###', getWNCodeForOutputPostprocessing(job, '    '))

        script = script.replace(
            '###DOWNLOADINPUTFILES###', getWNCodeForDownloadingInputFiles(job, '    '))

        if jobconfig.env:
            script = script.replace(
                '###APPLICATIONENVS###', repr(jobconfig.env))
        else:
            script = script.replace('###APPLICATIONENVS###', repr({}))

        script = script.replace('###WRAPPERLOG###', repr(wrapperlog))
        import inspect
        script = script.replace(
            '###INLINEMODULES###', inspect.getsource(Sandbox.WNSandbox))

        mon = job.getMonitoringService()

        self.monInfo = None

        # set the monitoring file by default to the stdout
        if isinstance(self.monInfo, dict):
            self.monInfo['remotefile'] = 'stdout'

        # try to print out the monitoring service information in debug mode
        try:
            logger.debug('job info of monitoring service: %s' %
                         str(self.monInfo))
        except:
            pass

#       prepare input/output sandboxes
        from Ganga.GPIDev.Lib.File import File
        from Ganga.Core.Sandbox.WNSandbox import PYTHON_DIR
        import inspect

        fileutils = File( inspect.getsourcefile(Ganga.Utility.files), subdir=PYTHON_DIR )
        packed_files = jobconfig.getSandboxFiles() + [ fileutils ]
        sandbox_files = job.createPackedInputSandbox(packed_files)

        # sandbox of child jobs should include master's sandbox
        sandbox_files.extend(master_job_sandbox)

        # check the input file size and pre-upload larger inputs to the iocache
        lfc_host = ''

        input_sandbox_uris = []
        input_sandbox_names = []

        ick = True

        max_prestaged_fsize = 0
        for f in sandbox_files:

            idx = self.__check_and_prestage_inputfile__(f)

            if not idx:
                logger.error('input sandbox preparation failed: %s' % f)
                ick = False
                break
            else:

                if idx['lfc_host']:
                    lfc_host = idx['lfc_host']

                if idx['remote']:
                    abspath = os.path.abspath(f)
                    fsize = os.path.getsize(abspath)

                    if fsize > max_prestaged_fsize:
                        max_prestaged_fsize = fsize

                    input_sandbox_uris.append(
                        idx['remote'][os.path.basename(f)])

                    input_sandbox_names.append(
                        os.path.basename(urlparse(f)[2]))

                if idx['local']:
                    input_sandbox_uris += idx['local']
                    input_sandbox_names.append(os.path.basename(f))

        if not ick:
            logger.error('stop job submission')
            return None

        # determin the lcg-cp timeout according to the max_prestaged_fsize
        # - using the assumption of 1 MB/sec.
        max_prestaged_fsize = 0
        lfc_host = ''
        transfer_timeout = config['SandboxTransferTimeout']
        predict_timeout = int(math.ceil(max_prestaged_fsize / 1000000.0))

        if predict_timeout > transfer_timeout:
            transfer_timeout = predict_timeout

        if transfer_timeout < 60:
            transfer_timeout = 60

        script = script.replace(
            '###TRANSFERTIMEOUT###', '%d' % transfer_timeout)

        # update the job wrapper with the inputsandbox list
        script = script.replace(
            '###INPUTSANDBOX###', repr({'remote': {}, 'local': input_sandbox_names}))

        # write out the job wrapper and put job wrapper into job's inputsandbox
        scriptPath = inpw.writefile(
            FileBuffer('__jobscript_%s__' % job.getFQID('.'), script), executable=1)
        input_sandbox = input_sandbox_uris + [scriptPath]

        for isb in input_sandbox:
            logger.debug('ISB URI: %s' % isb)

        # compose output sandbox to include by default the following files:
        # - gzipped stdout (transferred only when the JobLogHandler is WMS)
        # - gzipped stderr (transferred only when the JobLogHandler is WMS)
        # - __jobscript__.log (job wrapper's log)
        output_sandbox = [wrapperlog]

        from Ganga.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns
        for outputSandboxPattern in getOutputSandboxPatterns(job):
            output_sandbox.append(outputSandboxPattern)

        if config['JobLogHandler'] in ['WMS']:
            output_sandbox += ['stdout.gz', 'stderr.gz']

        if len(jobconfig.outputbox):
            output_sandbox += [Sandbox.OUTPUT_TARBALL_NAME]

        # compose ARC XRSL
        xrsl = {
            #'VirtualOrganisation' : config['VirtualOrganisation'],
            'executable': os.path.basename(scriptPath),
            'environment': {'GANGA_LCG_VO': config['VirtualOrganisation'], 'GANGA_LOG_HANDLER': config['JobLogHandler'], 'LFC_HOST': lfc_host},
            #'stdout'                : 'stdout',
            #'stderr'                : 'stderr',
            'inputFiles': input_sandbox,
            'outputFiles': output_sandbox,
            #'OutputSandboxBaseDestURI': 'gsiftp://localhost'
        }

        xrsl['environment'].update({'GANGA_LCG_CE': self.CE})
        #xrsl['Requirements'] = self.requirements.merge(jobconfig.requirements).convert()

        # if self.jobtype.upper() in ['NORMAL','MPICH']:
        #xrsl['JobType'] = self.jobtype.upper()
        # if self.jobtype.upper() == 'MPICH':
        #xrsl['Requirements'].append('(other.GlueCEInfoTotalCPUs >= NodeNumber)')
        # xrsl['Requirements'].append('Member("MPICH",other.GlueHostApplicationSoftwareRunTimeEnvironment)')
        #xrsl['NodeNumber'] = self.requirements.nodenumber
        # else:
        #    logger.warning('JobType "%s" not supported' % self.jobtype)
        #    return

#       additional settings from the job
        if jobconfig.env:
            xrsl['environment'].update(jobconfig.env)

        xrslText = Grid.expandxrsl(xrsl)
        
        # append any additional requirements from the requirements object
        xrslText += '\n'.join(self.requirements.other)

        logger.debug('subjob XRSL: %s' % xrslText)
        return inpw.writefile(FileBuffer('__xrslfile__', xrslText))

    def kill(self):
        '''Kill the job'''
        job = self.getJobObject()

        logger.info('Killing job %s' % job.getFQID('.'))

        if not self.id:
            logger.warning('Job %s is not running.' % job.getFQID('.'))
            return False

        return grids['GLITE'].arc_cancel([self.id])

    def master_kill(self):
        '''kill the master job to the grid'''

        job = self.getJobObject()

        if not job.master and len(job.subjobs) == 0:
            return IBackend.master_kill(self)
        elif job.master:
            return IBackend.master_kill(self)
        else:
            return self.master_bulk_kill()

    def master_bulk_kill(self):
        '''GLITE bulk resubmission'''

        job = self.getJobObject()

        # killing the individually re-submitted subjobs
        logger.debug('cancelling running/submitted subjobs.')

        # 1. collect job ids
        ids = []
        for sj in job.subjobs:
            if sj.status in ['submitted', 'running'] and sj.backend.id:
                ids.append(sj.backend.id)

        # 2. cancel the collected jobs
        ck = grids['GLITE'].arc_cancelMultiple(ids)
        if not ck:
            logger.warning('Job cancellation failed')
            return False
        else:
            for sj in job.subjobs:
                if sj.backend.id in ids:
                    sj.updateStatus('killed')

            return True

    def master_bulk_submit(self, rjobs, subjobconfigs, masterjobconfig):
        '''submit multiple subjobs in parallel, by default using 10 concurrent threads'''

        from Ganga.Utility.logic import implies
        assert(implies(rjobs, len(subjobconfigs) == len(rjobs)))

        # prepare the subjobs, jdl repository before bulk submission
        node_jdls = self.__mt_job_prepare__(
            rjobs, subjobconfigs, masterjobconfig)

        if not node_jdls:
            logger.error('Some jobs not successfully prepared')
            return False

        # set all subjobs to submitting status
        for sj in rjobs:
            sj.updateStatus('submitting')

        node_jids = self.__mt_bulk_submit__(node_jdls)

        status = False

        if node_jids:
            for sj in rjobs:
                if sj.id in node_jids.keys():
                    sj.backend.id = node_jids[sj.id]
                    sj.backend.CE = self.CE
                    sj.backend.actualCE = sj.backend.CE
                    sj.updateStatus('submitted')
                    sj.info.submit_counter += 1
                else:
                    logger.warning(
                        'subjob %s not successfully submitted' % sj.getFQID('.'))

            status = True

        return status

    def master_bulk_resubmit(self, rjobs):
        '''ARC bulk resubmission'''

        from Ganga.Utility.logging import log_user_exception

#        job = self.getJobObject()

        # compose master JDL for collection job
        node_jdls = {}
        for sj in rjobs:
            jdlpath = os.path.join(sj.inputdir, '__jdlfile__')
            node_jdls[sj.id] = jdlpath

        # set all subjobs to submitting status
        for sj in rjobs:
            sj.updateStatus('submitting')

        node_jids = self.__mt_bulk_submit__(node_jdls)

        status = False

        if node_jids:
            for sj in rjobs:
                if sj.id in node_jids.keys():
                    self.__refresh_jobinfo__(sj)
                    sj.backend.id = node_jids[sj.id]
                    sj.backend.CE = self.CE
                    sj.backend.actualCE = sj.backend.CE
                    sj.updateStatus('submitted')
                    sj.info.submit_counter += 1
                else:
                    logger.warning(
                        'subjob %s not successfully submitted' % sj.getFQID('.'))

            status = True

#            # set all subjobs to submitted status
#            # NOTE: this is just a workaround to avoid the unexpected transition
#            #       that turns the master job's status from 'submitted' to 'submitting'.
#            #       As this transition should be allowed to simulate a lock mechanism in Ganga 4, the workaround
#            #       is to set all subjobs' status to 'submitted' so that the transition can be avoided.
#            #       A more clear solution should be implemented with the lock mechanism introduced in Ganga 5.
#            for sj in rjobs:
#                sj.updateStatus('submitted')
#                sj.info.submit_counter += 1

        return status

    def master_submit(self, rjobs, subjobconfigs, masterjobconfig):
        '''Submit the master job to the grid'''

        profiler = ElapsedTimeProfiler(getLogger(name='Profile.LCG'))
        profiler.start()

        job = self.getJobObject()

        # finding ARC CE endpoint for job submission
        #allowed_celist = []
        # try:
        #    allowed_celist = self.requirements.getce()
        #    if not self.CE and allowed_celist:
        #        self.CE = allowed_celist[0]
        # except:
        #    logger.warning('ARC CE assigment from ARCRequirements failed.')

        # if self.CE and allowed_celist:
        #    if self.CE not in allowed_celist:
        #        logger.warning('submission to CE not allowed: %s, use %s instead' % ( self.CE, allowed_celist[0] ) )
        #        self.CE = allowed_celist[0]

        # use arc info to check for any endpoints recorded in the config file
        rc, output = grids['GLITE'].arc_info()

        if not self.CE and rc != 0:
            raise GangaException(
                "ARC CE endpoint not set and no default settings in '%s'. " % config['ArcConfigFile'])
        elif self.CE:
            logger.info('ARC CE endpoint set to: ' + str(self.CE))
        else:
            logger.info("Using ARC CE endpoints defined in '%s'" %
                        config['ArcConfigFile'])

        # delegate proxy to ARC CE
        # if not grids['GLITE'].arc_proxy_delegation(self.CE):
        #    logger.warning('proxy delegation to %s failed' % self.CE)

        # doing massive job preparation
        if len(job.subjobs) == 0:
            ick = IBackend.master_submit(
                self, rjobs, subjobconfigs, masterjobconfig)
        else:
            ick = self.master_bulk_submit(
                rjobs, subjobconfigs, masterjobconfig)

        profiler.check('==> master_submit() elapsed time')

        return ick

    def submit(self, subjobconfig, master_job_sandbox):
        '''Submit the job to the grid'''

        ick = False

        xrslpath = self.preparejob(subjobconfig, master_job_sandbox)

        if xrslpath:
            self.id = grids['GLITE'].arc_submit(
                xrslpath, self.CE, self.verbose)

            if self.id:
                self.actualCE = self.CE
                ick = True

        return ick

    def master_auto_resubmit(self, rjobs):
        """
        Resubmit each subjob individually as bulk resubmission will overwrite
        previous master job statuses
        """

        # check for master failure - in which case bulk resubmit
        mj = self._getParent()
        if mj.status == 'failed':
            return self.master_resubmit(rjobs)

        for j in rjobs:
            if not j.backend.master_resubmit([j]):
                return False

        return True

    def master_resubmit(self, rjobs):
        '''Resubmit the master job to the grid'''

        profiler = ElapsedTimeProfiler(getLogger(name='Profile.LCG'))
        profiler.start()

        job = self.getJobObject()

        ick = False

        # delegate proxy to ARC CE
        if not grids['GLITE'].arc_proxy_delegation(self.CE):
            logger.warning('proxy delegation to %s failed' % self.CE)

        if not job.master and len(job.subjobs) == 0:
            # case 1: master job normal resubmission
            logger.debug('rjobs: %s' % str(rjobs))
            logger.debug('mode: master job normal resubmission')
            ick = IBackend.master_resubmit(self, rjobs)

        elif job.master:
            # case 2: individual subjob resubmission
            logger.debug('mode: individual subjob resubmission')
            ick = IBackend.master_resubmit(self, rjobs)

        else:
            # case 3: master job bulk resubmission
            logger.debug('mode: master job resubmission')

            ick = self.master_bulk_resubmit(rjobs)
            if not ick:
                raise GangaException('ARC bulk submission failure')

        profiler.check('job re-submission elapsed time')

        return ick

    def resubmit(self):
        '''Resubmit the job'''

        ick = False

        job = self.getJobObject()

        jdlpath = job.getInputWorkspace().getPath("__jdlfile__")

        if jdlpath:
            self.id = grids['GLITE'].arc_submit(jdlpath, self.CE, self.verbose)

            if self.id:
                # refresh the lcg job information
                self.__refresh_jobinfo__(job)
                self.actualCE = self.CE
                ick = True

        return ick

    @staticmethod
    def updateMonitoringInformation(jobs):
        '''Monitoring loop for normal jobs'''

        import datetime

        backenddict = {}
        jobdict = {}
        for j in jobs:
            if j.backend.id and ((datetime.datetime.utcnow() - j.time.timestamps["submitted"]).seconds > config["ArcWaitTimeBeforeStartingMonitoring"]):
                jobdict[j.backend.id] = j
                backenddict[j.backend.actualCE] = j

        if len(jobdict.keys()) == 0:
            return

        jobInfoDict = grids['GLITE'].arc_status(
            jobdict.keys(), backenddict.keys())
        jidListForPurge = []

        # update job information for those available in jobInfoDict
        for id, info in jobInfoDict.items():

            if info:

                job = jobdict[id]

                if job.backend.actualCE != urlparse(id)[1].split(":")[0]:
                    job.backend.actualCE = urlparse(id)[1].split(":")[0]

                if job.backend.status != info['State']:

                    doStatusUpdate = True

                    # no need to update Ganga job status if backend status is
                    # not changed
                    if info['State'] == job.backend.status:
                        doStatusUpdate = False

                    # download output sandboxes if final status is reached
                    elif info['State'] in ['Finished', '(FINISHED)', 'Finished (FINISHED)']:

                        # grab output sandbox
                        if grids['GLITE'].arc_get_output(job.backend.id, job.getOutputWorkspace(create=True).getPath()):
                            (ick, app_exitcode) = grids['GLITE'].__get_app_exitcode__(
                                job.getOutputWorkspace(create=True).getPath())
                            job.backend.exitcode = app_exitcode

                            jidListForPurge.append(job.backend.id)

                        else:
                            logger.error(
                                'fail to download job output: %s' % jobdict[id].getFQID('.'))

                    if doStatusUpdate:
                        job.backend.status = info['State']
                        if 'Exit Code' in info:
                            try:
                                job.backend.exitcode_arc = int(
                                    info['Exit Code'])
                            except:
                                job.backend.exitcode_arc = 1

                        if 'Job Error' in info:
                            try:
                                job.backend.reason = info['Job Error']
                            except:
                                pass

                        job.backend.updateGangaJobStatus()
            else:
                logger.warning(
                    'fail to retrieve job informaton: %s' % jobdict[id].getFQID('.'))

        # purging the jobs the output has been fetched locally
        if jidListForPurge:
            if not grids['GLITE'].arc_purgeMultiple(jidListForPurge):
                logger.warning("Failed to purge all ARC jobs.")


    def updateGangaJobStatus(self):
        '''map backend job status to Ganga job status'''

        job = self.getJobObject()

        if self.status.startswith('Running') or self.status.startswith('Finishing'):
            job.updateStatus('running')
        elif self.status.startswith('Finished'):
            if job.backend.exitcode and job.backend.exitcode != 0:
                job.backend.reason = 'non-zero app. exit code: %s' % repr(
                    job.backend.exitcode)
                job.updateStatus('failed')
            elif job.backend.exitcode_arc and job.backend.exitcode_arc != 0:
                job.backend.reason = 'non-zero ARC job exit code: %s' % repr(
                    job.backend.exitcode_arc)
                job.updateStatus('failed')
            else:
                job.updateStatus('completed')

        elif self.status in ['DONE-FAILED', 'ABORTED', 'UNKNOWN', 'Failed']:
            job.updateStatus('failed')

        elif self.status in ['CANCELLED']:
            job.updateStatus('killed')

        elif self.status.startswith('Queuing'):
            pass

        else:
            logger.warning('Unexpected job status "%s"', self.status)

logger = getLogger()





###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: DefaultSplitter.py,v 1.2 2008-09-09 15:11:35 moscicki Exp $
###############################################################################

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Base.Proxy import addProxy, stripProxy
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem

from Ganga.Utility.logging import getLogger
logger = getLogger()


class DefaultSplitter(ISplitter):

    """
        The DefaultSplitter is assigned to all jobs by default and is intended to provide a single subjob for every job on submit.
        This has been implemented as it potentially simplifies the internal logic in job managment significantly.

        This splitter is not expected to be configurable or to split a dataset based upon any input.
        In order to do this please make use of another splitter.
    """
    _name = "DefaultSplitter"
    ## A dummy value is required to not get a bug in writing the object to an XML repo.
    ## The nature of the problem of writing an empty schema should probably be understood more correctly but is difficult to track down -rcurrie
    _schema = Schema(Version(1, 0), {'dummy_value': SimpleItem(defvalue=1, hidden=1, visitable=0, doc='the number of files per subjob', typelist=["int"])})

    def split(self, job):

        subjobs = []

        sj = self.createSubjob(job)

        subjobs.append(sj)

        return subjobs





from Ganga.Utility.Config import getConfig

configGaudi = getConfig('GAUDI')


def getpack(self, options=''):
    if self.newStyleApp is True:
        return getpack_cmake(self, options)
    else:
        return getpack_CMT(self, options)


def make(self, argument=None):
    if self.newStyleApp is True:
        return make_cmake(self, argument)
    else:
        return make_CMT(self, argument)


def getpack_cmake(appname):
    raise NotImplementedError


def make_cmake(self, argument=None):
    raise NotImplementedError


def getpack_CMT(self, options=''):
    """Performs a getpack on the package given within the environment
       of the application. The unix exit code is returned
    """
    command = 'getpack ' + options + '\n'
    if options == '':
        command = 'getpack -i'
    from GangaLHCb.Lib.Applications.CMTscript import CMTscript
    return CMTscript(self, command)


def make_CMT(self, argument=None):
    """Performs a CMT make on the application. The unix exit code is
       returned. Any arguments given are passed onto CMT as in
       dv.make('clean').
    """
    command = configGaudi['make_cmd']
    if argument:
        command += ' ' + argument
    from GangaLHCb.Lib.Applications.CMTscript import CMTscript
    return CMTscript(self, command)



#!/usr/bin/env python

from __future__ import print_function

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import sys
import time
import subprocess
import string
import shutil
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

commands = {'init': 'lhcb-proxy-init', 'info': 'dirac-proxy-info',
                    'destroy': 'grid-proxy-destroy'}

try:
    option = sys.argv[1]
    argv = sys.argv[2:]
except:
    print("No options passed, options are:")
    for k in commands:
        print(k)
    print("")
    e = Exception()
    raise e

global_env = None


def get_env():
    global global_env
    if global_env is None:
        global_env = {}
        for k, v in os.environ.iteritems():
            if not str(v).startswith('() {'):
                global_env[k] = v
            else:
                this_string = str(v).split('\n')
                final_str = ""
                for line in this_string:
                    final_str += str(os.path.expandvars(line)).strip()
                    if not final_str.endswith(';'):
                        final_str += " ;"
                    final_str += " "
                global_env[k] = final_str

    return global_env

if "GANGADIRACENVIRONMENT" not in get_env():
    e = Exception()
    e.args = ('DIRAC env cache file does not exist.',)
    raise e
temp_env = get_env()
dirac_env_cache_file = temp_env["GANGADIRACENVIRONMENT"]
if not os.path.exists(dirac_env_cache_file):
    e = Exception()
    e.args = ('DIRAC env cache file does not exist.',)
    raise e
env_file = open(dirac_env_cache_file)
for line in env_file.readlines():
    varval = line.strip().split('=')
    contents = ''.join(varval[1:])
    if not contents.startswith('() {'):
        global_env[varval[0]] = contents

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def get_stdout(cmd):
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=open('/dev/null'), env=get_env())
    proc.wait()
    if proc.poll() == 0:
        return proc.communicate()[0]
    else:
        return None


def exec_command(cmd, argv):
    #for arg in argv: cmd += ' %s' % arg
    command_list = [cmd]
    for arg in argv:
        command_list.append(arg)
    rc = subprocess.call(command_list, env=get_env())
    return rc


def get_proxy_file():
    stdout = get_stdout(['grid-proxy-info', '-path'])
    if not stdout:
        return ''
    return string.strip(stdout)


def get_proxy_cache_file():
    proxy_file = get_proxy_file()
    if not proxy_file:
        return ''
    return string.strip(proxy_file) + '.ganga-cache.py'


def get_proxy_timeleft():
    stdout = get_stdout(['grid-proxy-info', '-timeleft'])
    if stdout is None:
        return 0
    return int(string.strip(stdout))


def create_proxy_cache():
    proxy_timeleft = get_proxy_timeleft()
    now = time.time()
    proxy_info = get_stdout([commands['info']])
    if proxy_info is None:
        return False
    timeleft_str = None
    for line in proxy_info.splitlines():
        if line.find('timeleft') >= 0:
            timeleft_str = string.join(line.split(':')[1:], ':')
            break
    proxy_info = proxy_info.replace(timeleft_str, ' ###TIMELEFT###')
    cache_file = open(get_proxy_cache_file(), 'w')
    cache_file.write('proxy_cache_time = %f\n' % now)
    cache_file.write('proxy_timeleft_when_cached = %d\n' % proxy_timeleft)
    cache_file.write("proxy_info_stdout = '''\n%s'''\n" % proxy_info)
    cache_file.close()
    shutil.copymode(get_proxy_file(), get_proxy_cache_file())
    return True


def destroy_proxy_cache(): os.system('rm -f %s' % get_proxy_cache_file())


def get_file_age(file): return time.time() - os.path.getctime(file)


def get_proxy_info():
    loc = {}
    execfile(get_proxy_cache_file(), {}, loc)
    time_since_cache = time.time() - loc['proxy_cache_time']
    timeleft = int(loc['proxy_timeleft_when_cached'] - time_since_cache)
    if timeleft < 0:
        timeleft = 0
    hours = timeleft / 3600
    minutes = timeleft / 60 - hours * 60
    seconds = timeleft - (minutes * 60 + hours * 3600)
    timeleft_str = '%02d:%02d:%02d' % (hours, minutes, seconds)
    return loc['proxy_info_stdout'].replace('###TIMELEFT###', timeleft_str)


def proxy_cache_is_valid():
    proxy_cache_file = get_proxy_cache_file()
    if not proxy_cache_file:
        return False
    if not os.path.exists(proxy_cache_file):
        return False
    cache_age = get_file_age(proxy_cache_file)
    proxy_age = get_file_age(get_proxy_file())
    if cache_age > 300 or cache_age > proxy_age:
        return False
    return True

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
rc = 0
if option == 'init':
    destroy_proxy_cache()
    rc = exec_command(commands[option], argv)
    if rc == 0:
        create_proxy_cache()
elif option == 'info':
    rc = exec_command('grid-proxy-info', ['-exists'])
    if rc != 0:
        destroy_proxy_cache()
    else:
        display = True
        if not proxy_cache_is_valid():
            destroy_proxy_cache()
            if not create_proxy_cache():
                rc = 1
                display = False
        if display:
            print(get_proxy_info())
elif option == 'destroy':
    destroy_proxy_cache()
    rc = exec_command(commands[option], argv)
else:
    e = Exception()
    e.args = ('Error! Option "%s" no recognized.' % option,)
    raise e
#
sys.exit(rc)
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



import JobInfoParser

jlist = []
for j in jobs['DIANE_4']:
    jlist.append(j)

dataTable = JobInfoParser.getJobInfoTable(jlist,['backend','backend.actualCE','application','status','backend.status'])


from GangaPlotter import *
plotter = GangaPlotter()

plotter.piechart(dataTable,pltColId=4,pltTitle='Pie Chart of backend status')


### Users' Interface
from GangaPlotter import *
plotter = GangaPlotter()

# no dataproc
plotter.piechart(jobs['DIANE_4'],attr='backend.CE',title='Job distribution by queue',output='test.png')

# user defined dataproc
plotter.piechart(jobs['DIANE_4'],attr='backend.CE',title='Job distribution by CE',output='test.png',dataproc=lambda x:x.split(':2119/')[0])

# build in dataproc
plotter.piechart(jobs['DIANE_4'],attr='backend.CE',title='Job distribution by CE',output='test.png',attrext='by_ce')
plotter.piechart(jobs['DIANE_4'],attr='backend.CE',title='Job distribution by queue',output='test.png',attrext='by_queue')
plotter.piechart(jobs['DIANE_4'],attr='backend.CE',title='Job distribution by country',output='test.png',attrext='by_country')



from __future__ import absolute_import
from Ganga import GPI
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import addProxy
from Ganga.GPIDev.Base.Proxy import stripProxy
from .common import logger
from Ganga.Utility.ColourText import ANSIMarkup, overview_colours
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
import time

markup = ANSIMarkup()

########################################################################


class Task(GangaObject):

    """This is a Task without special properties"""
    _schema = Schema(Version(1, 0), {
        'transforms': ComponentItem('transforms', defvalue=[], sequence=1, copyable=1, doc='list of transforms'),
        'id': SimpleItem(defvalue=-1, protected=1, doc='ID of the Task', typelist=["int"]),
        'name': SimpleItem(defvalue='NewTask', copyable=1, doc='Name of the Task', typelist=["str"]),
        'comment': SimpleItem('', protected=0, doc='comment of the task', typelist=["str"]),
        'status': SimpleItem(defvalue='new', protected=1, doc='Status - new, running, pause or completed', typelist=["str"]),
        'float': SimpleItem(defvalue=0, copyable=1, doc='Number of Jobs run concurrently', typelist=["int"]),
        'resub_limit': SimpleItem(defvalue=0.9, copyable=1, doc='Resubmit only if the number of running jobs is less than "resub_limit" times the float. This makes the job table clearer, since more jobs can be submitted as subjobs.', typelist=["float"]),
        'creation_date': SimpleItem(defvalue="19700101", copyable=0, hidden=1, doc='Creation date of the task (used in dq2 datasets)', typelist=["str"]),
    })

    _category = 'tasks'
    _name = 'Task'
    _exportmethods = [
        # Settings
        'setBackend', 'setParameter', 'insertTransform', 'appendTransform', 'removeTransform',
        'check', 'run', 'pause', 'remove',  # Operations
        # Info
        'overview', 'info', 'n_all', 'n_status', 'help', 'getJobs', 'table',
        'float_all', 'run_all'  # Helper
    ]

    default_registry = "tasks"

# Special methods:
    def _auto__init__(self, registry=None):
        if registry is None:
            from Ganga.Core.GangaRepository import getRegistry
            registry = getRegistry(self.default_registry)
        # register the job (it will also commit it)
        # job gets its id now
        registry._add(self)
        self.creation_date = time.strftime('%Y%m%d%H%M%S')
        self.initialize()
        self.startup()
        self._setDirty()

    def initialize(self):
        pass

    def startup(self):
        """Startup function on Ganga startup"""
        for t in self.transforms:
            t.startup()

#   def _readonly(self):
#      """A task is read-only if the status is not new."""
#      if self.status == "new":
#         return 0
#      return 1

# Public methods:
#
# - remove() a task
# - clone() a task
# - check() a task (if updated)
# - run() a task to start processing
# - pause() to interrupt processing
# - setBackend(be) for all transforms
# - setParameter(myParam=True) for all transforms
# - insertTransform(id, tf) insert a new processing step
# - removeTransform(id) remove a processing step

    def remove(self, remove_jobs="do_nothing"):
        """Delete the task"""
        if not remove_jobs in [True, False]:
            logger.info("You want to remove the task %i named '%s'." % (self.id, self.name))
            logger.info("Since this operation cannot be easily undone, please call this command again:")
            logger.info(" * as tasks(%i).remove(remove_jobs=True) if you want to remove all associated jobs," % (self.id))
            logger.info(" * as tasks(%i).remove(remove_jobs=False) if you want to keep the jobs." % (self.id))
            return
        if remove_jobs:
            for j in GPI.jobs:
                try:
                    stid = j.application.tasks_id.split(":")
                    if int(stid[-2]) == self.id:
                        j.remove()
                except Exception as err:
                    logger.debug("Task remove_jobs task split Error!")
                    logger.debug("Error:\n%s" % str(err))
                    pass
        self._getRegistry()._remove(self)
        logger.info("Task #%s deleted" % self.id)

    def clone(self):
        c = super(Task, self).clone()
        for tf in c.transforms:
            tf.status = "new"
            # This is cleared separately since it is not in the schema
            tf._partition_apps = {}
        # self._getParent().register(c)
        c.check()
        return c

    def check(self):
        """This function is called by run() or manually by the user"""
        if self.status != "new":
            logger.error("The check() function may modify a task and can therefore only be called on new tasks!")
            return
        try:
            for t in self.transforms:
                t.check()
        finally:
            self.updateStatus()
        return True

    def run(self):
        """Confirms that this task is fully configured and ready to be run."""
        if self.status == "new":
            self.check()

        if self.status != "completed":
            if self.float == 0:
                logger.warning("The 'float', the number of jobs this task may run, is still zero. Type 'tasks(%i).float = 5' to allow this task to submit 5 jobs at a time" % self.id)
            try:
                for tf in self.transforms:
                    if tf.status != "completed":
                        tf.run(check=False)

            finally:
                self.updateStatus()
        else:
            logger.info("Task is already completed!")

    def pause(self):
        """Pause the task - the background thread will not submit new jobs from this task"""
        float_cache = self.float
        self.float = 0
        if self.status != "completed":
            for tf in self.transforms:
                tf.pause()
            self.status = "pause"
        else:
            logger.info("Transform is already completed!")
        self.float = float_cache

    def setBackend(self, backend):
        """Sets the backend on all transforms"""
        for tf in self.transforms:
            if backend is None:
                tf.backend = None
            else:
                tf.backend = stripProxy(backend).clone()

    def setParameter(self, **args):
        """Use: setParameter(processName="HWW") to set the processName in all applications to "HWW"
           Warns if applications are not affected because they lack the parameter"""
        for name, parm in args.iteritems():
            for tf in [t for t in self.transforms if t.application]:
                if name in tf.application.getNodeData():
                    addProxy(tf.application).__setattr__(name, parm)
                else:
                    logger.warning("Transform %i was not affected!", tf.name)

    def insertTransform(self, id, tf):
        """Insert transfrm tf before index id (counting from 0)"""
        if self.status != "new" and id < len(self.transforms):
            logger.error("You can only insert transforms at the end of the list. Only if a task is new it can be freely modified!")
            return
        # self.transforms.insert(id,tf.copy()) # this would be safer, but
        # breaks user exspectations
        # this means that t.insertTransform(0,t2.transforms[0]) will cause
        # Great Breakage
        self.transforms.insert(id, tf)

    def appendTransform(self, tf):
        """Append transform"""
        return self.insertTransform(len(self.transforms), tf)

    def removeTransform(self, id):
        """Remove the transform with the index id (counting from 0)"""
        if self.status != "new":
            logger.error("You can only remove transforms if the task is new!")
            return
        del self.transforms[id]

    def getJobs(self, transform=None, partition=None, only_master_jobs=True):
        """ Get the job slice of all jobs that process this task """
        if not partition is None:
            only_master_jobs = False
        jobslice = JobRegistrySlice("tasks(%i).getJobs(transform=%s, partition=%s, only_master_jobs=%s)" % (
            self.id, transform, partition, only_master_jobs))

        def addjob(j):
            if transform is None or partition is None or self.transforms[int(transform)]._app_partition[j.application.id] == partition:
                jobslice.objects[j.fqid] = stripProxy(j)

        for j in GPI.jobs:
            try:
                stid = j.application.tasks_id.split(":")
                if int(stid[-2]) == self.id and (transform is None or stid[-1] == str(transform)):
                    if j.subjobs and not only_master_jobs:
                        for sj in j.subjobs:
                            addjob(sj)
                    else:
                        addjob(j)
            except Exception as err:
                logger.debug("getJobs Error!!")
                logger.debug("Error:\n%s" % str(err))
                # print x
                pass
        return JobRegistrySliceProxy(jobslice)

# Internal methods
    def finaliseTransforms(self):
        """Check for any things needing doing after a transform has completed"""
        for t in self.transforms:
            t.finalise()

    def updateStatus(self):
        """Updates status based on transform status.
           Called from check() or if status of a transform changes"""
        # Calculate status from transform status:
        states = [tf.status for tf in self.transforms]
        if "running" in states and "pause" in states:
            new_status = "running/pause"
        elif "running" in states:
            new_status = "running"
        elif "pause" in states:
            new_status = "pause"
        elif "new" in states:
            new_status = "new"
        elif "completed" in states:
            new_status = "completed"
        else:
            new_status = "new"  # no tranforms
        # Handle status changes here:
        if self.status != new_status:
            if new_status == "running/pause":
                logger.info("Some Transforms of Task %i '%s' have been paused. Check tasks.table() for details!" % (
                    self.id, self.name))
            elif new_status == "completed":
                logger.warning("Task %i '%s' has completed!" %
                               (self.id, self.name))
            elif self.status == "completed":
                logger.warning(
                    "Task %i '%s' has been reopened!" % (self.id, self.name))
        self.status = new_status
        return self.status

    def submitJobs(self):
        """Submits as many jobs as necessary to maintain the float. Internal"""
        numjobs = 0
        if not self.status in ["running", "running/pause"]:
            return 0

        for i in range(len(self.transforms) - 1, -1, -1):
            tf = self.transforms[i]
            to_run = self.float - self.n_status("running")
            run = (self.resub_limit * self.float >= self.n_status("running"))
            if tf.status == "running" and to_run > 0 and run:
                numjobs += tf.submitJobs(to_run)
        return numjobs

    # Information methods
    def n_all(self):
        return sum([t.n_all() for t in self.transforms])

    def n_status(self, status):
        return sum([t.n_status(status) for t in self.transforms])

    def table(self):
        from Ganga.GPI import tasks
        tasks[self.id:self.id].table()

    def overview(self):
        """ Get an ascii art overview over task status. Can be overridden """
        logger.info("Colours: " + ", ".join([markup(key, overview_colours[key])
                                             for key in ["hold", "ready", "running", "completed", "attempted", "failed", "bad", "unknown"]]))
        logger.info(
            "Lists the partitions of events that are processed in one job, and the number of failures to process it.")
        logger.info("Format: (partition number)[:(number of failed attempts)]")
        logger.info('')
        for t in self.transforms:
            t.overview()

    def info(self):
        for t in self.transforms:
            t.info()

    def help(self):
        logger.info("This is a Task without special properties")

    # Helper methods
    def float_all(self):
        self.float = self.n_all()

    def run_all(self):
        self.float_all()
        self.run()



from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaLHCb.Lib.RTHandlers.LHCbGaudiRunTimeHandler import LHCbGaudiRunTimeHandler
#from GangaTest.Lib.TestSubmitter import TestSubmitter
try:
    ##    from GangaLHCb.Lib.DIRAC.GaudiDiracRTHandler import GaudiDiracRTHandler
    ##    from GangaLHCb.Lib.DIRAC.RootDiracRTHandler import RootDiracRTHandler
    from GangaLHCb.Lib.RTHandlers.LHCbGaudiDiracRunTimeHandler import LHCbGaudiDiracRunTimeHandler
    from GangaLHCb.Lib.RTHandlers.LHCbRootDiracRunTimeHandler import LHCbRootDiracRunTimeHandler
    loadRootHandler = True
except ImportError:
    loadRootHandler = False

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

from Ganga.GPI import *
import os.path
DAVINCI_VERSION = None


def getDiracAppPlatform():
    return 'x86_64-slc6-gcc48-opt'

# Add Local runtimehandler for Testsubmitter


def addLocalTestSubmitter():
    allHandlers.add('DaVinci', 'TestSubmitter', LHCbGaudiRunTimeHandler)


class DiracTestSubmitter(TestSubmitter):

    def __init__(self):
        super(DiracTestSubmitter, self).__init__()
        self._impl.diracOpts = ''
        self._impl.settings = {}

# Add Dirac runtimehandler for Testsubmitter


def addDiracTestSubmitter():
    allHandlers.add('DaVinci', 'TestSubmitter', LHCbGaudiDiracRunTimeHandler)
    allHandlers.add(
        'GaudiPython', 'TestSubmitter', LHCbGaudiDiracRunTimeHandler)
    if loadRootHandler:
        allHandlers.add('Root', 'TestSubmitter', LHCbRootDiracRunTimeHandler)

# Get Current test version


def getTestDaVinciVersion():
    d = {}
    # FIXME once config option is available
    siteInstallArea = '/afs/cern.ch/sw/ganga/install/TEST'
    execfile(
        os.path.join(siteInstallArea, 'LHCb', 'software', 'CURRENT'), d, d)
    return d['DAVINCI_VERSION']


def getTestDaVinciApplication():
    # FIXME once config option is available
    siteInstallArea = '/afs/cern.ch/sw/ganga/install/TEST'
    app = DaVinci()
    app.version = getTestDaVinciVersion()
    app.user_release_area = os.path.join(
        siteInstallArea, 'LHCb', 'software', 'install_area')
    hat = 'Tutorial'
    package = 'Analysis'
    tutorialversion = 'v6r5'
    app.masterpackage = package + ' ' + tutorialversion + ' ' + hat
    app.optsfile = [os.path.join(app.user_release_area,
                                 'DaVinci_' + str(app.version),
                                 hat, package, tutorialversion,
                                 'solutions', 'DaVinci2', 'DVTutorial_2.opts')]
    return app


def checkFileExists(file):
    import os.path
    return os.path.isfile(file)


def checkFolderExists(folder):
    import os.path
    return os.path.isdir(folder)

# check for a file in a tar file


def checkFileInTar(tf, file):
    import tarfile
    tar1 = tarfile.open(tf, "r")
    logger.info("%s files in %s" % (str(len(tar1.getnames())), tf))
    for fileName in tar1.getnames():
        if str(fileName[:2]) == str('./'):
            logger.info("found file: %s" % str(fileName[2:]))
            testFileName = str(fileName[2:])
        else:
            logger.info("found file: %s" % fileName)
            testFileName = fileName
        if file == testFileName:
            return True
        else:
            continue

    # return (file in tar1.getnames())
    return False

# check for a file in the inputsandbox


def checkFileInSandbox(job, file):
    import os.path
    masterTest = False
    subJobTest = False
    tarFile2 = os.path.join(
        job.inputdir, '_input_sandbox_%d_master.tgz' % job.id)
    tarFile1 = os.path.join(job.inputdir, '_input_sandbox_%d.tgz' % job.id)

    logger.info("Checking: %s and %s" % (tarFile1, tarFile2))

    if checkFileExists(tarFile2):
        masterTest = checkFileInTar(tarFile2, file)
    elif checkFileExists(tarFile1):
        subJobTest = checkFileInTar(tarFile1, file)
    else:
        pass

    if masterTest is True:
        return True
    elif subJobTest is True:
        return True
    else:
        return False




def getEnvironment(c = None):
    import PACKAGE
    PACKAGE.standardSetup()
    return {}

def loadPlugins( config = None ):
    import Lib.TestApplication
    import Lib.TestSubmitter # TestSubmitter, TestSplitter here
    import Lib.TestRobot
    import Lib.TestObjects
    import Framework.runner





from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy
from ND280Unit_CSVEvtList import ND280Unit_CSVEvtList
from modules.ND280Dataset.ND280Dataset import ND280LocalDataset, ND280DCacheDataset
from modules.ND280Splitter.ND280Splitter import splitCSVFile
import os

class ND280Transform_CSVEvtList(ITransform):
   _schema = Schema(Version(1,0), dict(ITransform._schema.datadict.items() + {
        'nbevents' : SimpleItem(defvalue=-1,doc='The number of events for each unit'),
    }.items()))

   _category = 'transforms'
   _name = 'ND280Transform_CSVEvtList'
   _exportmethods = ITransform._exportmethods + [ ]

   def __init__(self):
      super(ND280Transform_CSVEvtList,self).__init__()

   def createUnits(self):
      """Create new units if required given the inputdata"""
      
      # call parent for chaining
      super(ND280Transform_CSVEvtList,self).createUnits()
      
      # Look at the application schema and check if there is a csvfile variable
      try:
        csvfile = self.application.csvfile
      except AttributeError:
        logger.error('This application doesn\'t contain a csvfile variable. Use another Transform !')
        return

      subsets = splitCSVFile(self.application.csvfile, self.nbevents)

      for s,sub in enumerate(subsets):
        
        # check if this data is being run over by checking all the names listed
        ok = False
        for unit in self.units:
          if unit.subpartid == s:
            ok = True

        if ok:
          continue
        # new unit required for this dataset
        unit = ND280Unit_CSVEvtList()
        unit.name = "Unit %d" % len(self.units)
        unit.subpartid = s
        unit.eventswanted = sub
        unit.inputdata = self.inputdata[0]
        self.addUnitToTRF( unit )


   def createChainUnit( self, parent_units, use_copy_output = True ):
      """Create a chained unit using the output data from the given units"""

      # check all parent units for copy_output
      copy_output_ok = True
      for parent in parent_units:
         if not parent.copy_output:
            copy_output_ok = False

      # all parent units must be completed so the outputfiles are filled correctly
      for parent in parent_units:
         if parent.status != "completed":
               return None

      if not use_copy_output or not copy_output_ok:
         unit = ND280Unit_CSVEvtList()
         unit.inputdata = ND280LocalDataset()
         for parent in parent_units:
            # loop over the output files and add them to the ND280LocalDataset - THIS MIGHT NEED SOME WORK!
            job = GPI.jobs(parent.active_job_ids[0])
            for f in job.outputfiles:
               # should check for different file types and add them as appropriate to the dataset
               # self.inputdata (== TaskChainInput).include/exclude_file_mask could help with this
               # This will be A LOT easier with Ganga 6.1 as you can easily map outputfiles -> inputfiles!
               unit.inputdata.names.append( os.path.join( job.outputdir, f.namePattern ) )
      else:

         unit = ND280Unit_CSVEvtList()
         unit.inputdata = ND280LocalDataset()

         for parent in parent_units:
            # unit needs to have completed and downloaded before we can get file list
            if parent.status != "completed":
               return None

            # we should be OK so copy all output to the dataset
            for f in parent.copy_output.files:
               unit.inputdata.names.append( os.path.join( parent.copy_output.local_location, f ) )
         
      return unit
   



from __future__ import print_function

##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: __init__.py,v 1.7 2009-06-25 13:30:00 moscicki Exp $
##########################################################################

#
# Thin wrapper for logging subsystem:
#  - makes standard python 2.3 logging available in python 2.2
#  - integrates loggers with Ganga configuration subsystem in a most covinient way
#  - extra reporting utilities:
#     log_user_exception
#
# Design principles:
#  - getLogger() returns the standard logger object for the current execution context
#     i.e. with the name which indicates the current package or module name
#  - all loggers are automatically configured according to this modules config dictionary (see below)
#  - special functions:
#       - log_user_exception() allows to format nicely exception messages

logging = None
handlers = None

def _get_logging():

    global logging

    if logging is None:
        import logging as _logging
        logging = _logging

    return logging

def _get_handlers():

    global handlers

    if handlers is None:
        import logging.handlers as _handlers

        handlers = _handlers

    return handlers

import sys

# logger configuration
# settings for new loggers may be added here at will (for example read
# from the config file)
import Ganga.Utility.Config

import Ganga.Utility.ColourText as ColourText

import threading

from Ganga.Utility.Config import getConfig
config = getConfig("Logging")

# initialize the root logger for the logger created directly in python
# executable scripts which have no name starting by "Ganga."
# By default everything goes to stdout

_hasInit = False

if not _hasInit:
    _get_logging().basicConfig(stream=sys.stdout)
    _hasIinit = True

_formats = {
    'DEBUG': '%(asctime)s "%(filename)s":%(funcName)-10s at %(lineno)d, %(threadName)s: %(levelname)-8s %(message)s',
    'VERBOSE': '%(asctime)s %(name)-35s: %(levelname)-8s %(message)s',
    'NORMAL': '%(name)-35s: %(levelname)-8s %(message)s',
    'TERSE': 'Ganga: %(levelname)-8s %(message)s'
}

requires_shutdown = False

private_logger = None  # private logger of this module

# main logger corresponds to the root of the hierarchy
main_logger = _get_logging().getLogger()

# this is the handler used to print on screen directly
direct_screen_handler = main_logger.handlers[0]  # get default StreamHandler

# if defined this is the handler used for caching background messages at
# interactive prompt
cached_screen_handler = None

# this is the handler currenty in use by main_logger (either direct_screen_handler or cached_screen_handler)
# or it may be overriden by bootstrap() to be arbitrary handler
default_handler = direct_screen_handler

# if defined this is ADDITIONAL handler that is used for the logfile
file_handler = None


# FIXME: this should be probably an option in the config
# if defined, global level (string) overrides anything which is in config
_global_level = None

# all loggers which are used by all modules
_allLoggers = {}

# use this function to get new loggers into your packages
# if you do not provide the name then logger will detect your package name
# if you specify the modulename as a string then it will be appended to the package name
# if you specify the modulename==1 then your module name will be guessed and appended to the package name
# the guessing algorithm may be modified by passing the frame object (to emulate a different physical location of the logger)
# this is only useful for special usage such as IBackend base class
def getLogger(name=None, modulename=None, frame=None):
    return _getLogger(name, modulename, frame=frame)

class ColourFormatter(_get_logging().Formatter, object):

    def __init__(self, *args, **kwds):
        _get_logging().Formatter.__init__(self, *args, **kwds)
        fg = ColourText.Foreground()
        fx = ColourText.Effects()
        self.colours = {_get_logging().INFO: fx.normal,
                        _get_logging().WARNING: fg.orange,
                        _get_logging().ERROR: fg.red,
                        _get_logging().CRITICAL: fg.red,
                        _get_logging().DEBUG: fx.normal}
        self.markup = ColourText.ANSIMarkup()

    def format(self, record):
        try:
            s = super(ColourFormatter, self).format(record)
        except TypeError:
            print("%s" % str(record))
            return None
        if record.levelno in self.colours.keys():
            code = self.colours[record.levelno]
            return self.markup(s, code)
        else:
            return s

    def setColour(self, yes):
        if yes:
            self.markup = ColourText.ANSIMarkup()
        else:
            self.markup = ColourText.NoMarkup()

def _set_formatter(handler):
    if config['_customFormat'] != "":
        for k in _formats.keys():
            _formats[k] = config['_customFormat']

    formatter = ColourFormatter(_formats[config['_format']])
    formatter.setColour(config['_colour'])
    handler.setFormatter(formatter)

def _make_file_handler(logfile, logfile_size):
    import os.path
    logfile = os.path.expanduser(logfile)
    global file_handler
    if logfile:
        try:
            new_file_handler = _get_handlers().RotatingFileHandler(
                logfile, maxBytes=logfile_size, backupCount=1)
        except IOError as x:
            private_logger.error('Cannot open the log file: %s', str(x))
            return
        # remove old handler if exists
        # print 'removing old file handler',file_handler
        # print 'installing new file handler',new_file_handler
        if file_handler:
            main_logger.removeHandler(file_handler)
            file_handler.flush()
            file_handler.close()
            # this is required to properly remove the file handler from the logging system
            # otherwise I/O Error at shutdown
            try:
                # WARNING: this relies on the implementation details of the
                # logging module
                del _get_logging()._handlers[file_handler]
            except KeyError:
                # don't complain if the handler was correctly unregistered by
                # the logging system
                pass

        new_file_handler.setFormatter(_get_logging().Formatter(_formats['VERBOSE']))
        main_logger.addHandler(new_file_handler)
        file_handler = new_file_handler

# set the loglevel for a logger to a given string value (example: "DEBUG")
def _set_log_level(logger, value):

    global _global_level

    if _global_level is not None:
        value = _global_level

    # convert a string "DEBUG" into enum object logging.DEBUG
    def _string2level(name):
        if hasattr(logging, name):
            return getattr(logging, name)

    try:
        logger.setLevel(_string2level(value))
        #if value != 'DEBUG':
        #    return value
        #    #logging._srcfile = None
        #    #logging.logThreads = 0
        #    #logging.logThreads = 0
        return value
    except AttributeError as err:
        logger.error('Attribute Error: %s', str(err))
        logger.warning('possible configuration error: invalid level value (%s), using default level', value)
        return None


# reflect all user changes immediately
def post_config_handler(opt, value):

    if config is not None and '_customFormat' in config and config['_customFormat'] != "":
        for k in _formats.keys():
            _formats[k] = config['_customFormat']

    if config is not None:
        _format, colour = config['_format'], config['_colour']

    #print("")# rcurrie - This strangely seems to trick the logging into working when the _customFormat is used on first start

    if opt in ['_format', '_customFormat']:
        badConfig = False
        if opt == "_customFormat":
            if config is not None:
                value = config['_format']
            else:
                value = None
        if _formats is not None and value in _formats:
            _format = _formats[value]
        else:
            if private_logger is not None:
                private_logger.error('illegal name of format string (%s), possible values: %s' % (str(value), _formats.keys()))
            return

    if opt == '_colour':
        colour = value

    if opt in ['_format', '_colour', '_customFormat']:
        fmt = ColourFormatter(_format)
        fmt.setColour(colour)
        direct_screen_handler.setFormatter(fmt)
        return

    if config is not None:
        logfile, logfile_size = config['_logfile'], config['_logfile_size']

    if opt in ['_logfile', '_logfile_size']:
        global file_handler
        _make_file_handler(logfile, logfile_size)
        return

    # FIXME: has no effect at runtime, should raise a ConfigError
    if opt == '_interactive_cache':
        return

    # set the logger level
    if private_logger is not None:
        private_logger.info('setting loglevel: %s %s', opt, value)

    if _set_log_level is not None and getLogger is not None:
        _set_log_level(getLogger(opt), value)


config.attachUserHandler(None, post_config_handler)
config.attachSessionHandler(None, post_config_handler)

lookup_frame_names = {}

def _guess_module_logger_name(modulename, frame=None):
    # print " _guess_module_logger_name",modulename
    # find the filename of the calling module
    import sys
    import os.path
    if frame is None:
        # assuming 2 nested calls to the module boundary!
        frame = sys._getframe(3)
    else:
        print('using frame from the caller')

    global lookup_frame_names


    this__file__ = None
    if '__file__' in frame.f_globals.keys():
        this__file__ = frame.f_globals['__file__']
        if this__file__ in lookup_frame_names:
            del frame
            return lookup_frame_names[this_file]
        else:
            should_store = True
    else:
        should_store = False

    # accessing __file__ from globals() is much more reliable than
    # f_code.co_filename (name = os.path.normcase(frame.f_code.co_filename))
    if this__file__ is not None:
        name = os.path.realpath(os.path.abspath(this__file__))
    else:
        # no file associated with the frame (e.g. interactive prompt, exec
        # statement)
        name = '_program_'

    # print " _guess_module_logger_name",name
    del frame

    # if private_logger:
    #    private_logger.debug('searching for package matching calling module co_filename= %s',str(name))

    # sometimes the filename is an absolute path, try to find a relative module path from the PYTHONPATH
    # and remove the trailing path -> the result will be used as the logger
    # name

    from Ganga.Utility.files import remove_prefix

    name = remove_prefix(name, sys.path)

    def remove_tail(s, tail):
        idx = s.rfind(tail)
        if idx != -1:
            return s[:idx]
        return s

    # get rid of trailing .py  .pyc .pyo
    name = remove_tail(name, '.py')

    # replace slashes with dots
    name = name.replace(os.sep, '.')

    # return full module name
    if modulename == 1:
        return name

    # remove module name
    name = remove_tail(name, '.')

    if name == 'ganga':  # interactive IPython session
        name = "Ganga.GPI"

    # return package name
    if not modulename:
        return name

    return_name = name + '.' + modulename

    if should_store is True:
        lookup_frame_names[this__file__] = return_name

    # return custom module name
    return return_name

_MemHandler = _get_handlers().MemoryHandler

# Caching will not be done for messages which are generated by the main thread.
class FlushedMemoryHandler(_MemHandler):

    def __init__(self, *args, **kwds):
       _MemHandler.__init__(self, *args, **kwds)

    def shouldFlush(self, record):
        return threading.currentThread().getName().find("GANGA_Update_Thread") == -1 or \
                _MemHandler.shouldFlush(self, record)
        #        #threading.currentThread().getName().find("Main") == -1
        #return False

def enableCaching():
    """
    Enable caching of log messages at interactive prompt. In the interactive IPython session, the messages from monitoring
    loop will be cached until the next prompt. In non-interactive sessions no caching is required.
    """

    if not config['_interactive_cache']:
        return

    private_logger.debug('CACHING ENABLED')
    global default_handler, cached_screen_handler
    main_logger.removeHandler(default_handler)
    cached_screen_handler = FlushedMemoryHandler(1000, target=direct_screen_handler)
    default_handler = cached_screen_handler
    main_logger.addHandler(default_handler)


def _getLogger(name=None, modulename=None, _roothandler=0, handler=None, frame=None):

    if handler is None:
        handler = default_handler

    requested_name = name

    if name is None:
        name = _guess_module_logger_name(modulename, frame=frame)

    if name.split('.')[0] != 'Ganga' and name != 'Ganga':
        name = 'Ganga.' + name

    ## Reduce verbosity on startup
    #if private_logger:
    #    private_logger.debug('getLogger: effective_name=%s original_name=%s', name, requested_name)

    if name in _allLoggers:
        return _allLoggers[name]
        # print 'reusing existing logger: ',name
    else:

        logger = _get_logging().getLogger(name)
        _allLoggers[name] = logger
        #logger.setLevel( _get_logging().CRITICAL )

        if name in config:
            thisConfig = config[name]
            _set_log_level(logger, thisConfig)

        ## Reduce verbosity on startup
        #if private_logger:
        #    private_logger.debug('created logger %s in %s mode', name, _get_logging().getLevelName(logger.getEffectiveLevel()))
        #    private_logger.debug('applied %s format string to %s', config['_format'], name)

        # print '------------>',logger
        #logger.critical('initialization of logger for module %s',name)
        # print '------------> should have printed the message'

        return logger


# this bootstrap method should be called very early in the bootstrap phase of the entire system to enable
# the log level configuration to take effect...
# the optional handler overrides the default handler and is used for GUI environment
# NOTE: the additional message buffering handler is available only for the
# default handler

def bootstrap(internal=False, handler=None):

    global private_logger, main_logger

    if internal == True and private_logger is not None:
        return

    private_logger = getLogger('Ganga.Utility.logging')
    #_set_log_level(private_logger, 'CRITICAL')
    #main_logger = _getLogger('Ganga',_roothandler=1,handler=handler)

    private_logger.debug('bootstrap')

    if internal:
        _make_file_handler(config['_logfile'], config['_logfile_size'])
    else:
        # override main_logger handler
        global default_handler
        if handler is not None and handler is not default_handler:
            main_logger.removeHandler(default_handler)
            main_logger.addHandler(handler)
            default_handler = handler

    _set_formatter(default_handler)
    # main_logger.propagate = 0 # do not propagate messages upwards...
    # main_logger.addHandler(default_handler)
    if file_handler:
        main_logger.addHandler(file_handler)

    opts = filter(lambda o: o.find('Ganga') == 0, config)
    for opt in opts:

        # should reconfigure Ganga and private logger according to the config
        # file contents

        # _get_logging().getLevelName(getLogger(opt).getEffectiveLevel()))
        msg = 'logging %s in %s mode' % (opt, config[opt])
        if internal:
            private_logger.debug(msg)
        else:
            private_logger.info(msg)

        _set_log_level(getLogger(opt), config[opt])

    if not internal:
        class NoErrorFilter(_get_logging().Filter):
            """
            A filter which only allow messages which are WARNING or lower to be logged
            """
            def filter(self, record):
                return record.levelno <= 30

        # Make the default handler not print ERROR and CRITICAL
        direct_screen_handler.addFilter(NoErrorFilter())

        # Add a new handler for ERROR and CRITICAL which prints to stderr
        error_logger = _get_logging().StreamHandler(sys.stderr)
        error_logger.setLevel(_get_logging().ERROR)
        _set_formatter(error_logger)
        main_logger.addHandler(error_logger)

    global requires_shutdown
    requires_shutdown = True

    private_logger.debug('end of bootstrap')


def final_shutdown():
    private_logger.debug('shutting down logsystem')
    _get_logging().shutdown()


# do the initial bootstrap automatically at import -> this will bootstrap
# the main and private loggers
bootstrap(internal=True)

# force all loggers to use the same level


def force_global_level(level):
    if level is not None:
        for l in _allLoggers:
            _set_log_level(_allLoggers[l], level)
        global _global_level
        _global_level = level


def log_user_exception(logger=None, debug=False):
    import traceback
    import cStringIO
    buf = cStringIO.StringIO()
    traceback.print_exc(file=buf)
    banner = 10 * '-' + ' error in user/extension code ' + 10 * '-'
    if not logger:
        logger = private_logger
    logmethod = logger.warning
    if debug:
        logmethod = logger.debug
    logmethod(banner)
    for line in buf.getvalue().splitlines():
        logmethod(line)
    logmethod('-' * len(banner))


def log_unknown_exception():
    """
    This is convenince function which is used to track down what exceptions
    are being caught by ambiguous 'except' clauses.
    It should only be called from within an exception handler.
    """
    tb_logger = getLogger('Ganga.Utility.logging.log_unknown_exception')

    # Fetch the place from where this function was called to locate the bare
    # except
    from inspect import getframeinfo, stack, getsourcefile
    caller = getframeinfo(stack()[1][0])

    tb_logger.debug('Bare except clause triggered {0}:{1}'.format(caller.filename, caller.lineno))
    tb_logger.debug('Exception caught:', exc_info=True)


# extra imports for more convenient work with the logging module
getLevelName = _get_logging().getLevelName



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: DQ2SandboxCache.py,v 1.9 2009-06-09 07:54:42 hclee Exp $
###############################################################################
#
# ATLAS/ARDA
#
# Date:   January 2007
import os
import os.path
import re
import tempfile
import shutil
from types import *
from urlparse import urlparse

from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *

from Ganga.Utility.logging import getLogger
from Ganga.Utility.GridShell import getShell 

from Ganga.Lib.LCG.GridSandboxCache import GridSandboxCache, GridFileIndex
from Ganga.Lib.LCG.Utility import *

dq2_sandbox_cache_schema_datadict = GridSandboxCache._schema.inherit_copy().datadict
dq2_file_index_schema_datadict    = GridFileIndex._schema.inherit_copy().datadict

class DQ2FileIndex(GridFileIndex):
    """
    Data object containing DQ2 file index information. 
    
    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com
    """
    dq2_file_index_schema_datadict.update({
        'dataset'        : SimpleItem(defvalue='', doc='the DQ2 dataset name'),
        'site'           : SimpleItem(defvalue='', doc='the DQ2 site id')
        } )

    _schema   = Schema( Version(1,0), dq2_file_index_schema_datadict )
    _category = 'GridFileIndex'
    _name = 'DQ2FileIndex'

    def __init__(self):
        super(DQ2FileIndex,self).__init__()

#    def __init__(self, surl, dataset, site, name, md5sum):
#
#        self.__init__()
#
#        self.id     = surl
#        self.name   = name
#        self.md5sum = md5sum
#        self.attributes['dataset'] = dataset
#        self.attributes['site']    = site

class DQ2SandboxCache(GridSandboxCache):

    '''
    Helper class for upladong/downloading/deleting files/datasets using DQ2 libraries.

    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com
    '''

    dq2_sandbox_cache_schema_datadict.update({
        'setup'          : SimpleItem(defvalue='/afs/cern.ch/atlas/offline/external/GRID/ddm/DQ2Clients/setup.sh', doc='the DQ2 setup script'),
        'local_site_id'  : SimpleItem(defvalue='CERN-PROD_SCRATCHDISK', copyable=1, doc='the DQ2 local site id'),
        'dataset_name'   : SimpleItem(defvalue='', copyable=0, doc='the DQ2 dataset name')
        } )

    _schema   = Schema( Version(1,0), dq2_sandbox_cache_schema_datadict )
    _category = 'GridSandboxCache'
    _name = 'DQ2SandboxCache'
 
    logger = getLogger()

    def __init__(self):
        super(DQ2SandboxCache,self).__init__()
        self.protocol = 'dq2'

    def impl_upload(self, files=[], opts=''):
        """
        Uploads multiple files to a remote grid storage.
        """

        shell = getShell(self.middleware)

        ## exclude the Ganga-owned external package for LFC python binding
        pythonpaths = []
        for path in shell.env['PYTHONPATH'].split(':'):
            if not re.match('.*\/external\/lfc\/.*', path):
                pythonpaths.append(path)
        shell.env['PYTHONPATH'] = ':'.join(pythonpaths)

        if self.local_site_id:
            shell.env['DQ2_LOCAL_SITE_ID'] = self.local_site_id

        self.logger.debug('DQ2_LOCAL_SITE_ID: %s' % shell.env['DQ2_LOCAL_SITE_ID'])

        # check dq2 existence
        rc,output,m = self.__cmd_retry_loop__(shell, 'source %s 2>&1 > /dev/null; which dq2-put' % self.setup, 1)
        if (rc != 0) or (not output):
            self.logger.error('dq2 client command not found. Please check DQ2 client installation.')
            return None

        # move all local files in the temporary directory
        src_dir = tempfile.mkdtemp(prefix='_ganga_tmp_')
        finfo   = {}
        myuuid  = get_uuid()

        try:
            for f in files:
                path  = os.path.dirname( urlparse(f)[2] )
                name  = os.path.basename( urlparse(f)[2] )

                tmp_fname = name + '_' + myuuid
                tmp_fpath = os.path.join(src_dir, tmp_fname )

                shutil.copy(os.path.join(path, name), tmp_fpath)
                md5sum = get_md5sum(os.path.join(path, name), ignoreGzipTimestamp=True)
                fsize  = os.path.getsize( tmp_fpath )
                finfo[tmp_fname] = {}
                finfo[tmp_fname]['md5sum'] = md5sum
                finfo[tmp_fname]['fsize']  = fsize
                finfo[tmp_fname]['local_fpath']  = urlparse(f)[2]
                finfo[tmp_fname]['surl'] = ''

            # compose dq2-put command
            cmd = 'source %s 2>&1 > /dev/null; export VOMS_PROXY_INFO_DONT_VERIFY_AC=1; dq2-put -a -d -C ' % (self.setup)

            if self.local_site_id:
                cmd += '-L %s ' % self.local_site_id

            cmd += '-s %s ' % src_dir
            cmd += '-f %s ' % ','.join(finfo.keys())
            cmd += '%s ' % self.dataset_name

            # run dq2-put
            rc,output,m = self.__cmd_retry_loop__(shell, cmd, self.max_try)

            self.logger.debug('%d %s' % (rc, output))

            # run dq2-ls to query the uploaded files
            # together with local file information, creates GridFileIndex objects
            file_idx = []
            if rc == 0:
                # compose dq2-ls command
                cmd = 'source %s 2>&1 > /dev/null; export VOMS_PROXY_INFO_DONT_VERIFY_AC=1; dq2-ls -L %s -f -p %s' % (self.setup, self.local_site_id, self.dataset_name)
                rc,output,m = self.__cmd_retry_loop__(shell, cmd, self.max_try)

                if rc == 0:

                    lines = map( lambda x:x.strip(), output.split() )

                    for l in lines:
                        for f in finfo.keys():
                            if l.find('srm://') == 0 and l.find(f) > 0:
                                finfo[f]['surl'] = l
                                break

                    for f in finfo.keys():
                        name   = os.path.basename( finfo[f]['local_fpath'] )
                        surl   = finfo[f]['surl']
                        md5sum = finfo[f]['md5sum']

                        ## create DQ2FileIndex of the uploaded file
                        fidx = DQ2FileIndex()
                        fidx.id = surl
                        fidx.name = name
                        fidx.dataset = self.dataset_name
                        fidx.site = self.local_site_id
                        fidx.md5sum = md5sum

                        file_idx.append( fidx )
        finally:
            ## removing the temporary directory in any case
            shutil.rmtree(src_dir)

        return file_idx     

#    def impl_download(self, files={}, dest_dir=None, opts=''):
#        """
#        Downloads multiple files from remote grid storages to 
#        a local directory.
#        """
#
#        return False

#    def impl_delete(self, files=[], opts=''):
#        """
#        Deletes whole dataset from the sit
#        """
#
#        ## ToDo: a better implementation
#        ##   1. get list of SURLs from the dataset
#        ##   2. if the given files is a subset - remove files with lcg-del and call dq2-check-replica-consistency
#        ##   3. if the given files covers all files in dataset - call dq2-delete-replicas
#
#
#        isDone = False
#
#        shell = getShell(self.middleware)
#
#        cmd = 'source %s; dq2-delete-replicas -d %s %s' % (self.setup, self.dataset_name, self.local_site_id)
#
#        rc,output,m = self.__cmd_retry_loop__(shell, cmd, self.max_try)
#
#        logger.debug(output)
#
#        if rc == 0:
#            isDone = True
#
#        return files



"""
Configuration options for the remote autonomous testing
of new releases of Ganga
"""

import Ganga.Utility.Config

config=Ganga.Utility.Config.makeConfig('TestRobot','Parameters for the autonomous independent testing of Ganga')

config.addOption('ReleasePath','http://cern/ch/ganga/download/','Default URL to check for Ganga releases')
config.addOption('InstallPath','','default install path for Ganga scratch testing')
config.addOption('JobDir','','dir to put job data in')
config.addOption('PluginsRequired','GangaTest','Add additional plugins if non-vanilla tests are required')
config.addOption('TestPairs',[['Ganga/test', 'Local', 'local'], ['GangaLHCb/test', 'Local', 'local']],'These are the tests to perform, the backend, and the configuration')
config.addOption('ReleaseNumber','','Blank default release number - to be set in session')
config.addOption('JobTimeOut',1800,'time out for the pre-release tests in SECONDS')
config.addOption('VersionTime','None','Time of upload of last pre-release tested')
config.addOption('VersionNumber','None','Last Version tested')
config.addOption('SleepTime','600','Time to wait between checking of new releases')
config.addOption('EmailOnStartup','False','If you require the autonomous process to email you on startup following a fail')
config.addOption('LastCheckedTime','0','Last Checked Time of the VERSIONS.txt')
config.addOption('PublishPath','~/public_html/testrobot','Directory where test results should be made available as a .tgz file')
config.addOption('Site','Imperial','Name of site where test is carried out')



#exec python -x "$0" "$@"
import zlib,sys

#adler starting value is _not_ 0L
adler=1
filename=sys.argv[1]

try:
    openFile = open(filename, 'rb')
    
    for line in openFile:
        adler=zlib.adler32(line, adler)
        
except:
    raise Exception('Could not get checksum of %s'%filename)

openFile.close()

#backflip on 32bit
if adler < 0:
    adler = adler + 2**32
    
print str('%08x'%adler) #return as padded hexified string

